import numpy as np
import scipy.sparse as sp
import matplotlib.pyplot as plt

from .. import Problem
from .. import Utils
from .. import Mesh
from .. import Props


def lengthInCell(O, D, x, y, plotIt=False):

    maxD = np.sqrt(np.sum(D ** 2))
    D = D / maxD

    def dist(a):
        return O + a * D

    if plotIt:
        plt.plot(x[[0, 0, 1, 1, 0]], y[[0, 1, 1, 0, 0]], "b")
        plt.plot(O[0], O[1], "rs")
        d = np.r_[0, maxD]
        plt.plot(O[0] + d * D[0], O[1] + d * D[1], "k-")

    alp = np.array([-np.inf, np.inf, -np.inf, np.inf])

    if np.abs(D[0]) > 0:
        alp[0] = (x[0] - O[0]) / D[0]
        alp[1] = (x[1] - O[0]) / D[0]
        if plotIt:
            plt.plot(dist(alp[0])[0], dist(alp[0])[1], "mo")
            plt.plot(dist(alp[1])[0], dist(alp[1])[1], "mo")

    if np.abs(D[1]) > 0:
        alp[2] = (y[0] - O[1]) / D[1]
        alp[3] = (y[1] - O[1]) / D[1]
        if plotIt:
            plt.plot(dist(alp[2])[0], dist(alp[2])[1], "go")
            plt.plot(dist(alp[3])[0], dist(alp[3])[1], "go")

    midAlp = np.array(sorted(alp)[1:3])
    midAlp[midAlp < 0] = 0
    midAlp[midAlp > maxD] = maxD
    midPoint = dist(np.mean(midAlp))

    #     print alp, midAlp, midPoint

    if (
        midPoint[0] >= x[0]
        and midPoint[0] <= x[1]
        and midPoint[1] >= y[0]
        and midPoint[1] <= y[1]
    ):
        vec = dist(midAlp[0]) - dist(midAlp[1])
        if plotIt:
            c = np.c_[dist(midAlp[0]), dist(midAlp[1])]
            plt.plot(c[0, :], c[1, :], "r", lw=2)
        return np.sqrt(np.sum(vec ** 2))

    return None


def lineintegral(M, Tx, Rx):
    O, D = Tx, Rx - Tx
    I, J, V = [], [], []
    for i in range(M.nCx):
        for j in range(M.nCy):
            x = M.vectorNx[[i, i + 1]]
            y = M.vectorNy[[j, j + 1]]
            v = lengthInCell(O, D, x, y)
            if v is not None:
                I += [i]
                J += [j]
                V += [v]
    inds = Utils.sub2ind(M.vnC, np.array([I, J]).T)
    return inds, V


class StraightRayProblem(Problem.LinearProblem):

    slowness, slownessMap, slownessDeriv = Props.Invertible("Slowness model (1/v)")

    @property
    def A(self):
        if getattr(self, "_A", None) is not None:
            return self._A

        self._A = sp.lil_matrix((self.survey.nD, self.mesh.nC))
        row = 0
        for tx in self.survey.txList:
            for rx in tx.rxList:
                for loc_i in range(rx.locs.shape[0]):
                    inds, V = lineintegral(self.mesh, tx.loc, rx.locs[loc_i, :])
                    self._A[inds * 0 + row, inds] = V
                    row += 1

        return self._A

    def fields(self, m):
        self.model = m
        return self.A * self.slowness

    def Jvec(self, m, v, f=None):
        self.model = m
        # mt = self.model.transformDeriv
        # return self.A * ( mt * v )
        return self.A * self.slownessDeriv * v

    def Jtvec(self, m, v, f=None):
        self.model = m
        # mt = self.model.transformDeriv
        # return mt.T * ( self.A.T * v )
        return self.slownessDeriv.T * self.A.T * v



import SimPEG
import numpy as np
from SimPEG.Utils import Zero, closestPoints


class BaseRx(SimPEG.Survey.BaseTimeRx):
    locs = None
    rxType = None

    knownRxTypes = {
        "phi": ["phi", None],
        "ex": ["e", "x"],
        "ey": ["e", "y"],
        "ez": ["e", "z"],
        "jx": ["j", "x"],
        "jy": ["j", "y"],
        "jz": ["j", "z"],
    }

    def __init__(self, locs, times, rxType, **kwargs):
        SimPEG.Survey.BaseTimeRx.__init__(self, locs, times, rxType, **kwargs)

    @property
    def projField(self):
        """Field Type projection (e.g. e b ...)"""
        return self.knownRxTypes[self.rxType][0]

    def projGLoc(self, f):
        """Grid Location projection (e.g. Ex Fy ...)"""
        comp = self.knownRxTypes[self.rxType][1]
        if comp is not None:
            return f._GLoc(self.rxType) + comp
        return f._GLoc(self.rxType)

    def getTimeP(self, timesall):
        """
            Returns the time projection matrix.

            .. note::

                This is not stored in memory, but is created on demand.
        """
        time_inds = np.in1d(timesall, self.times)
        return time_inds

    def evalDeriv(self, src, mesh, f, v, adjoint=False):
        P = self.getP(mesh, self.projGLoc(f))
        if not adjoint:
            return P * v
        elif adjoint:
            return P.T * v


# DC.Rx.Dipole(locs)
class Dipole(BaseRx):

    rxgeom = "dipole"

    def __init__(self, locsM, locsN, times, rxType="phi", **kwargs):
        assert locsM.shape == locsN.shape, "locsM and locsN need to be the same size"
        if np.array_equal(locsM, locsN):
            self.rxgeom = "pole"
            locs = [locsM]
        else:
            locs = [locsM, locsN]
        # We may not need this ...
        BaseRx.__init__(self, locs, times, rxType)

    @property
    def nD(self):
        """Number of data in the receiver."""
        # return self.locs[0].shape[0] * len(self.times)
        return self.locs[0].shape[0]

    @property
    def nRx(self):
        """Number of data in the receiver."""
        return self.locs[0].shape[0]

        # Not sure why ...
        # return int(self.locs[0].size / 2)

    def getP(self, mesh, Gloc):

        if mesh in self._Ps:
            return self._Ps[mesh]

        if self.rxgeom == "dipole":
            P0 = mesh.getInterpolationMat(self.locs[0], Gloc)
            P1 = mesh.getInterpolationMat(self.locs[1], Gloc)
            P = P0 - P1
        elif self.rxgeom == "pole":
            P = mesh.getInterpolationMat(self.locs[0], Gloc)

        if self.storeProjections:
            self._Ps[mesh] = P

        return P



"""
`printinfo` -- Tools to print date, time, and version information
=================================================================

Print or return date, time, and package version information in any environment
(Jupyter notebook, IPython console, Python console, QT console), either as
html-table (notebook) or as plain text (anywhere).

This script is an adapted version of ``empymod.printversions``
(https://empymod.github.io), which itself was heavily inspired by

- ``ipynbtools.py`` from https://github.com/qutip, and
- ``watermark.py`` from https://github.com/rasbt/watermark,

Always shown are the OS, number of CPU(s), ``numpy``, ``scipy``, ``SimPEG``,
``cython``, ``properties``, ``vectormath``, ``discretize``, ``pymatsolver``,
``sys.version``, and time/date.

Additionally shown are, if they can be imported, ``IPython``, ``matplotlib``,
and ``ipywidgets``. It also shows MKL information, if available.

All modules provided in ``add_pckg`` are also shown. They have to be imported
before ``versions`` is called.

"""

# Mandatory modules
import sys
import time
import numpy
import scipy
import textwrap
import platform
import multiprocessing

# Required packages
import SimPEG
import cython
import properties
import vectormath
import discretize
import pymatsolver

# Optional modules
try:
    import IPython
    from IPython.display import HTML, Pretty
except ImportError:
    IPython = False
try:
    import matplotlib
except ImportError:
    matplotlib = False
try:
    import ipywidgets
except ImportError:
    ipywidgets = False
try:
    import numexpr
except ImportError:
    numexpr = False
try:
    import mkl
except ImportError:
    mkl = False

# Get mkl info from numexpr or mkl, if available
if mkl:
    mklinfo = mkl.get_version_string()
elif numexpr:
    mklinfo = numexpr.get_vml_version()
else:
    mklinfo = False

__all__ = ["versions", "versions_html", "versions_text"]


def versions(mode="print", add_pckg=None, ncol=4):
    """Return date, time, and version information.

    Print or return date, time, and package version information in any
    environment (Jupyter notebook, IPython console, Python console, QT
    console), either as html-table (notebook) or as plain text (anywhere).

    This script is an adapted version of ``empymod.printversions``,
    (https://empymod.github.io) which itself was heavily inspired by

        - ipynbtools.py from qutip https://github.com/qutip
        - watermark.py from https://github.com/rasbt/watermark

    This is a wrapper for ``versions_html`` and ``versions_text``.

    **Parameters**

    mode : string, optional; {<'print'>, 'HTML', 'Pretty', 'plain', 'html'}
        Defaults to 'print':
            - 'print': Prints text-version to stdout, nothing returned.
            - 'HTML': Returns html-version as IPython.display.HTML(html).
            - 'html': Returns html-version as plain text.
            - 'Pretty': Returns text-version as IPython.display.Pretty(text).
            - 'plain': Returns text-version as plain text.

        'HTML' and 'Pretty' require IPython.

    add_pckg : packages, optional
        Package or list of packages to add to output information (must be
        imported beforehand).

    ncol : int, optional
        Number of package-columns in html table; only has effect if
        ``mode='HTML'`` or ``mode='html'``. Defaults to 3.


    **Returns**

    Depending on ``mode`` (HTML-instance; plain text; html as plain text; or
    nothing, only printing to stdout).


    **Examples**

    >>> import pytest
    >>> import dateutil
    >>> from .. import versions
    >>> versions()                 # Default values
    >>> versions('plain', pytest)  # Provide additional package
    >>> versions('HTML', [pytest, dateutil], ncol=5)  # HTML

    """
    if mode == "html":
        return versions_html(add_pckg, ncol)
    elif mode == "plain":
        return versions_text(add_pckg)
    elif mode == "Pretty" and IPython:
        return Pretty(versions_text(add_pckg))
    elif mode == "HTML" and IPython:
        return HTML(versions_html(add_pckg, ncol))
    else:
        print(versions_text(add_pckg))


def versions_html(add_pckg=None, ncol=4):
    """HTML version.

    See ``versions`` for details.
    """

    # Check ncol
    ncol = int(ncol)

    # Define html-styles
    border = "border: 2px solid #fff;'"

    def colspan(html, txt, ncol, nrow):
        """Print txt in a row spanning whole table."""
        html += "  <tr>\n"
        html += "     <td style='text-align: center; "
        if nrow == 0:
            html += "font-weight: bold; font-size: 1.2em; "
        elif nrow % 2 == 0:
            html += "background-color: #ddd;"
        html += border + " colspan='"
        html += str(2 * ncol) + "'>%s</td>\n" % txt
        html += "  </tr>\n"
        return html

    def cols(html, version, name, ncol, i):
        """Print package information in two cells."""

        # Check if we have to start a new row
        if i > 0 and i % ncol == 0:
            html += "  </tr>\n"
            html += "  <tr>\n"

        html += "    <td style='text-align: right; background-color: #ccc; "
        html += border + ">%s</td>\n" % version

        html += "    <td style='text-align: left; "
        html += border + ">%s</td>\n" % name

        return html, i + 1

    # Start html-table
    html = "<table style='border: 3px solid #ddd;'>\n"

    # Date and time info as title
    html = colspan(html, time.strftime("%a %b %d %H:%M:%S %Y %Z"), ncol, 0)

    # OS and CPUs
    html += "  <tr>\n"
    html, i = cols(html, platform.system(), "OS", ncol, 0)
    html, i = cols(html, multiprocessing.cpu_count(), "CPU(s)", ncol, i)

    # Loop over packages
    for pckg in _get_packages(add_pckg):
        html, i = cols(html, pckg.__version__, pckg.__name__, ncol, i)
    # Fill up the row
    while i % ncol != 0:
        html += "    <td style= " + border + "></td>\n"
        html += "    <td style= " + border + "></td>\n"
        i += 1
    # Finish row
    html += "  </tr>\n"

    # sys.version
    html = colspan(html, sys.version, ncol, 1)

    # mkl version
    if mklinfo:
        html = colspan(html, mklinfo, ncol, 2)

    # Finish table
    html += "</table>"

    return html


def versions_text(add_pckg=None):
    """Plain-text version.

    See ``versions`` for details.
    """

    # Width for text-version
    n = 54
    text = "\n" + "".join(n * ["-"]) + "\n"

    # Date and time info as title
    text += time.strftime("  %a %b %d %H:%M:%S %Y %Z\n\n")

    # OS and CPUs
    text += f"{platform.system():>15}" + " : OS\n"
    text += f"{multiprocessing.cpu_count():>15}" + " : CPU(s)\n"

    # Loop over packages
    for pckg in _get_packages(add_pckg):
        text += f"{pckg.__version__:>15} : {pckg.__name__}\n"

    # sys.version
    text += "\n"
    for txt in textwrap.wrap(sys.version, n - 4):
        text += "  " + txt + "\n"

    # mkl version
    if mklinfo:
        text += "\n"
        for txt in textwrap.wrap(mklinfo, n - 4):
            text += "  " + txt + "\n"

    # Finish
    text += "".join(n * ["-"])
    return text


def _get_packages(add_pckg):
    """Create list of packages."""

    # Mandatory packages
    pckgs = [
        numpy,
        scipy,
        SimPEG,
        cython,
        properties,
        vectormath,
        discretize,
        pymatsolver,
    ]

    # Optional packages
    for module in [IPython, ipywidgets, matplotlib]:
        if module:
            pckgs += [module]

    # Cast and add add_pckg
    if add_pckg is not None:

        # Cast add_pckg
        if isinstance(add_pckg, tuple):
            add_pckg = list(add_pckg)

        if not isinstance(add_pckg, list):
            add_pckg = [
                add_pckg,
            ]

        # Add add_pckg
        pckgs += add_pckg

    return pckgs



import numpy as np
from scipy.interpolate import interp1d
from scipy.constants import mu_0


def TriangleFun(time, ta, tb):
    """
        Triangular Waveform
        * time: 1D array for time
        * ta: time at peak
        * tb: time at step-off
    """
    out = np.zeros(time.size)
    out[time <= ta] = 1 / ta * time[time <= ta]
    out[(time > ta) & (time < tb)] = (
        -1 / (tb - ta) * (time[(time > ta) & (time < tb)] - tb)
    )
    return out


def TriangleFunDeriv(time, ta, tb):
    """
        Derivative of Triangular Waveform
    """
    out = np.zeros(time.size)
    out[time <= ta] = 1 / ta
    out[(time > ta) & (time < tb)] = -1 / (tb - ta)
    return out


def SineFun(time, ta):
    """
        Sine Waveform
        * time: 1D array for time
        * ta: Pulse Period
    """
    out = np.zeros(time.size)
    out[time <= ta] = np.sin(1.0 / ta * np.pi * time[time <= ta])

    return out


def SineFunDeriv(time, ta):
    """
        Derivative of Sine Waveform
    """
    out = np.zeros(time.size)
    out[time <= ta] = 1.0 / ta * np.pi * np.cos(1.0 / ta * np.pi * time[time <= ta])
    return out


def VTEMFun(time, ta, tb, a):
    """
        VTEM Waveform
        * time: 1D array for time
        * ta: time at peak of exponential part
        * tb: time at step-off
    """
    out = np.zeros(time.size)
    out[time <= ta] = (1 - np.exp(-a * time[time <= ta] / ta)) / (1 - np.exp(-a))
    out[(time > ta) & (time < tb)] = (
        -1 / (tb - ta) * (time[(time > ta) & (time < tb)] - tb)
    )
    return out


def CausalConv(array1, array2, time):
    """
        Evaluate convolution for two causal functions.
        Input

        * array1: array for \\\\(\\\\ f_1(t)\\\\)
        * array2: array for \\\\(\\\\ f_2(t)\\\\)
        * time: array for time

        .. math::

            Out(t) = \\int_{0}^{t} f_1(a) f_2(t-a) da

    """

    if array1.shape == array2.shape == time.shape:
        out = np.convolve(array1, array2)
        # print time[1]-time[0]
        return out[0 : np.size(time)] * (time[1] - time[0])
    else:
        print("Give me same size of 1D arrays!!")


def RectFun(time, ta, tb):
    """

        Rectangular Waveform

        * time: 1D array for time
        * ta: time for transition from (+) to (-)
        * tb: time at step-off

        .. math::

            I(t) = 1, 0 < t \\le t_a

            I(t) = -1, t_a < t < t_b

            I(t) = 0, t \\le t_a \\ \\text{or}  \\ t \\ge t_b

    """
    out = np.zeros(time.size)
    out[time <= ta] = 1
    out[(time > ta) & (time < tb)] = -1
    return out


def CenDiff(f, tin):
    """
        Evaluating central difference of given array (f)
        and provide funtion handle for interpolation
    """
    dfdt = mu_0 * np.diff(f, n=1) / np.diff(tin, n=1)
    tm = np.diff(tin, n=1) * 0.5 + tin[:-1]
    Diffun = interp1d(tm, dfdt)
    return Diffun



import numpy as np
import scipy.sparse as sp
from scipy import constants
from .. import Utils, Props


def _get_projections(u):
    """Get the projections for each domain in the pressure head (u)"""
    nP = len(u)
    bools = u >= 0
    ind_p = np.where(bools)[0]
    ind_n = np.where(~bools)[0]
    P_p = sp.csr_matrix((np.ones(len(ind_p)), (ind_p, ind_p)), shape=(nP, nP))
    P_n = sp.csr_matrix((np.ones(len(ind_n)), (ind_n, ind_n)), shape=(nP, nP))
    return P_p, P_n


def _partition_args(mesh, Hcond, Theta, hcond_args, theta_args, **kwargs):

    hcond_params = {k: kwargs[k] for k in kwargs if k in hcond_args}
    theta_params = {k: kwargs[k] for k in kwargs if k in theta_args}

    other_params = {k: kwargs[k] for k in kwargs if k not in hcond_args + theta_args}

    if len(other_params) > 0:
        raise Exception(f"Unknown parameters: {other_params}")

    hcond = Hcond(mesh, **hcond_params)
    theta = Theta(mesh, **theta_params)

    return hcond, theta


class NonLinearModel(Props.HasModel):
    """A non linear model that has dependence on the fields and a model"""

    counter = None  #: A SimPEG.Utils.Counter object
    mesh = None  #: A SimPEG Mesh

    def __init__(self, mesh, **kwargs):
        self.mesh = mesh
        super().__init__(**kwargs)

    @property
    def nP(self):
        """Number of parameters in the model."""
        return self.mesh.nC


class BaseWaterRetention(NonLinearModel):
    def plot(self, ax=None):
        import matplotlib.pyplot as plt

        if ax is None:
            plt.figure()
            ax = plt.subplot(111)

        self.validate()

        h = -np.logspace(-2, 3, 1000)
        ax.semilogx(-h, self(h))
        ax.set_title("Water retention curve")
        ax.set_xlabel(r"Soil water potential, $- \psi$")
        ax.set_ylabel("Water content, $\\theta$")


class BaseHydraulicConductivity(NonLinearModel):
    def plot(self, ax=None):
        import matplotlib.pyplot as plt

        if ax is None:
            plt.figure()
            ax = plt.subplot(111)

        self.validate()

        h = -np.logspace(-2, 3, 1000)
        ax.loglog(-h, self(h))
        ax.set_title("Hydraulic conductivity function")
        ax.set_xlabel(r"Soil water potential, $- \psi$")
        ax.set_ylabel("Hydraulic conductivity, $K$")


class Haverkamp_theta(BaseWaterRetention):

    theta_r, theta_rMap, theta_rDeriv = Props.Invertible(
        "residual water content [L3L-3]", default=0.075
    )

    theta_s, theta_sMap, theta_sDeriv = Props.Invertible(
        "saturated water content [L3L-3]", default=0.287
    )

    alpha, alphaMap, alphaDeriv = Props.Invertible("", default=1.611e06)

    beta, betaMap, betaDeriv = Props.Invertible("", default=3.96)

    def _get_params(self):
        return self.theta_r, self.theta_s, self.alpha, self.beta

    def __call__(self, u):
        theta_r, theta_s, alpha, beta = self._get_params()

        f = alpha * (theta_s - theta_r) / (alpha + abs(u) ** beta) + theta_r

        if np.isscalar(theta_s):
            f[u >= 0] = theta_s
        else:
            f[u >= 0] = theta_s[u >= 0]
        return f

    def derivM(self, u):
        """derivative with respect to m

        .. code::

            import sympy as sy

            alpha, u, beta, theta_r, theta_s = sy.symbols(
                'alpha u beta theta_r theta_s', real=True
            )

            f_n = (
                alpha *
                (theta_s - theta_r) /
                (alpha + abs(u)**beta) +
                theta_r
            )
        """
        return (
            self._derivTheta_r(u)
            + self._derivTheta_s(u)
            + self._derivAlpha(u)
            + self._derivBeta(u)
        )

    def _derivTheta_r(self, u):
        if self.theta_rMap is None:
            return Utils.Zero()
        theta_r, theta_s, alpha, beta = self._get_params()
        ddm = -alpha / (alpha + abs(u) ** beta) + 1
        ddm[u >= 0] = 0
        dT = Utils.sdiag(ddm) * self.theta_rDeriv
        return dT

    def _derivTheta_s(self, u):
        if self.theta_sMap is None:
            return Utils.Zero()
        theta_r, theta_s, alpha, beta = self._get_params()
        P_p, P_n = _get_projections(u)  # Compute the positive/negative domains
        dT_p = P_p * self.theta_sDeriv
        dT_n = P_n * Utils.sdiag(alpha / (alpha + abs(u) ** beta)) * self.theta_sDeriv
        return dT_p + dT_n

    def _derivAlpha(self, u):
        if self.alphaMap is None:
            return Utils.Zero()
        theta_r, theta_s, alpha, beta = self._get_params()
        ddm = -alpha * (-theta_r + theta_s) / (alpha + abs(u) ** beta) ** 2 + (
            -theta_r + theta_s
        ) / (alpha + abs(u) ** beta)
        ddm[u >= 0] = 0
        dA = Utils.sdiag(ddm) * self.alphaDeriv
        return dA

    def _derivBeta(self, u):
        if self.betaMap is None:
            return Utils.Zero()
        theta_r, theta_s, alpha, beta = self._get_params()
        ddm = (
            -alpha
            * (-theta_r + theta_s)
            * np.log(abs(u))
            * abs(u) ** beta
            / (alpha + abs(u) ** beta) ** 2
        )
        ddm[u >= 0] = 0
        dN = Utils.sdiag(ddm) * self.betaDeriv
        return dN

    def derivU(self, u):
        theta_r, theta_s, alpha, beta = self._get_params()

        g = (
            alpha
            * ((theta_s - theta_r) / (alpha + abs(u) ** beta) ** 2)
            * (-beta * abs(u) ** (beta - 1) * np.sign(u))
        )
        g[u >= 0] = 0
        g = Utils.sdiag(g)
        return g


class Haverkamp_k(BaseHydraulicConductivity):

    Ks, KsMap, KsDeriv = Props.Invertible(
        "Saturated hydraulic conductivity", default=9.44e-03
    )

    A, AMap, ADeriv = Props.Invertible("fitting parameter", default=1.175e06)

    gamma, gammaMap, gammaDeriv = Props.Invertible("fitting parameter", default=4.74)

    def _get_params(self):
        return self.Ks, self.A, self.gamma

    def __call__(self, u):
        Ks, A, gamma = self._get_params()
        P_p, P_n = _get_projections(u)  # Compute the positive/negative domains
        f_p = P_p * np.ones(len(u)) * Ks  # ensures scalar Ks works
        f_n = P_n * Ks * A / (A + abs(u) ** gamma)
        return f_p + f_n

    def derivU(self, u):
        Ks, A, gamma = self._get_params()
        g = -(Ks * A * gamma * abs(u) ** (gamma - 1) * np.sign(u)) / (
            (A + abs(u) ** gamma) ** 2
        )
        g[u >= 0] = 0
        return Utils.sdiag(g)

    def derivM(self, u):
        return self._derivKs(u) + self._derivA(u) + self._derivGamma(u)

    def _derivKs(self, u):
        if self.KsMap is None:
            return Utils.Zero()

        Ks, A, gamma = self._get_params()
        P_p, P_n = _get_projections(u)  # Compute the positive/negative domains

        dKs_dm_p = P_p * self.KsDeriv
        dKs_dm_n = P_n * Utils.sdiag(A / (A + abs(u) ** gamma)) * self.KsDeriv
        return dKs_dm_p + dKs_dm_n

    def _derivA(self, u):
        if self.AMap is None:
            return Utils.Zero()
        Ks, A, gamma = self._get_params()
        ddm = Ks / (A + abs(u) ** gamma) - Ks * A / (A + abs(u) ** gamma) ** 2
        ddm[u >= 0] = 0
        dA_dm = Utils.sdiag(ddm) * self.ADeriv
        return dA_dm

    def _derivGamma(self, u):
        if self.gammaMap is None:
            return Utils.Zero()
        Ks, A, gamma = self._get_params()
        ddm = -(A * Ks * np.log(abs(u)) * abs(u) ** gamma) / (A + abs(u) ** gamma) ** 2
        ddm[u >= 0] = 0
        dGamma_dm = Utils.sdiag(ddm) * self.gammaDeriv
        return dGamma_dm


def haverkamp(mesh, **kwargs):
    return _partition_args(
        mesh,
        Haverkamp_k,
        Haverkamp_theta,
        ["Ks", "A", "gamma"],
        ["alpha", "beta", "theta_r", "theta_s"],
        **kwargs,
    )


class HaverkampParams:
    """Holds some default parameterizations for the Haverkamp model."""

    @property
    def celia1990(self):
        """Parameters used in:

            Celia, Michael A., Efthimios T. Bouloutas, and Rebecca L. Zarba.
            "A general mass-conservative numerical solution for the unsaturated
            flow equation." Water Resources Research 26.7 (1990): 1483-1496.
        """
        return {
            "alpha": 1.611e06,
            "beta": 3.96,
            "theta_r": 0.075,
            "theta_s": 0.287,
            "Ks": 9.44e-03,
            "A": 1.175e06,
            "gamma": 4.74,
        }


class Vangenuchten_theta(BaseWaterRetention):

    theta_r, theta_rMap, theta_rDeriv = Props.Invertible(
        "residual water content [L3L-3]", default=0.078
    )

    theta_s, theta_sMap, theta_sDeriv = Props.Invertible(
        "saturated water content [L3L-3]", default=0.430
    )

    n, nMap, nDeriv = Props.Invertible(
        "measure of the pore-size distribution, >1", default=1.56
    )

    alpha, alphaMap, alphaDeriv = Props.Invertible(
        "related to the inverse of the air entry suction [L-1], >0", default=0.036
    )

    def _get_params(self):
        return self.theta_r, self.theta_s, self.alpha, self.n

    def __call__(self, u):
        theta_r, theta_s, alpha, n = self._get_params()
        f = (theta_s - theta_r) / (
            (1.0 + abs(alpha * u) ** n) ** (1.0 - 1.0 / n)
        ) + theta_r
        if np.isscalar(theta_s):
            f[u >= 0] = theta_s
        else:
            f[u >= 0] = theta_s[u >= 0]

        return f

    def derivM(self, u):
        """derivative with respect to m

        .. code::

            import sympy as sy

            alpha, u, n, I, Ks, theta_r, theta_s = sy.symbols(
                'alpha u n I Ks theta_r theta_s', real=True
            )

            m = 1.0 - 1.0/n
            theta_e = 1.0 / ((1.0 + sy.functions.Abs(alpha * u) ** n) ** m)

            f_n = (
                (
                    theta_s - theta_r
                ) /
                (
                    (1.0 + abs(alpha * u)**n) ** (1.0 - 1.0 / n)
                ) +
                theta_r
            )
        """
        return (
            self._derivTheta_r(u)
            + self._derivTheta_s(u)
            + self._derivN(u)
            + self._derivAlpha(u)
        )

    def _derivTheta_r(self, u):
        if self.theta_rMap is None:
            return Utils.Zero()
        theta_r, theta_s, alpha, n = self._get_params()
        ddm = -((abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n)) + 1
        ddm[u >= 0] = 0
        dT = Utils.sdiag(ddm) * self.theta_rDeriv
        return dT

    def _derivTheta_s(self, u):
        if self.theta_sMap is None:
            return Utils.Zero()
        theta_r, theta_s, alpha, n = self._get_params()
        P_p, P_n = _get_projections(u)  # Compute the positive/negative domains
        dT_p = P_p * self.theta_sDeriv
        dT_n = (
            P_n
            * Utils.sdiag((abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n))
            * self.theta_sDeriv
        )
        return dT_p + dT_n

    def _derivN(self, u):
        if self.nMap is None:
            return Utils.Zero()
        theta_r, theta_s, alpha, n = self._get_params()
        ddm = (
            (-theta_r + theta_s)
            * (
                (-1.0 + 1.0 / n)
                * np.log(abs(alpha * u))
                * abs(alpha * u) ** n
                / (abs(alpha * u) ** n + 1.0)
                - 1.0 * np.log(abs(alpha * u) ** n + 1.0) / n ** 2
            )
            * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n)
        )
        ddm[u >= 0] = 0
        dN = Utils.sdiag(ddm) * self.nDeriv
        return dN

    def _derivAlpha(self, u):
        if self.alphaMap is None:
            return Utils.Zero()
        theta_r, theta_s, alpha, n = self._get_params()
        ddm = (
            n
            * u
            * (-1.0 + 1.0 / n)
            * (-theta_r + theta_s)
            * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n)
            * abs(alpha * u) ** n
            * np.sign(alpha * u)
            / ((abs(alpha * u) ** n + 1.0) * abs(alpha * u))
        )
        ddm[u >= 0] = 0
        dA = Utils.sdiag(ddm) * self.alphaDeriv
        return dA

    def derivU(self, u):
        theta_r, theta_s, alpha, n = self._get_params()
        g = (
            -alpha
            * n
            * abs(alpha * u) ** (n - 1)
            * np.sign(alpha * u)
            * (1.0 / n - 1)
            * (theta_r - theta_s)
            * (abs(alpha * u) ** n + 1) ** (1.0 / n - 2)
        )
        g[u >= 0] = 0
        g = Utils.sdiag(g)
        return g


class Vangenuchten_k(BaseHydraulicConductivity):

    Ks, KsMap, KsDeriv = Props.Invertible(
        "Saturated hydraulic conductivity", default=24.96
    )

    I, IMap, IDeriv = Props.Invertible("", default=0.5)

    n, nMap, nDeriv = Props.Invertible(
        "measure of the pore-size distribution, >1", default=1.56
    )

    alpha, alphaMap, alphaDeriv = Props.Invertible(
        "related to the inverse of the air entry suction [L-1], >0", default=0.036
    )

    def _get_params(self):
        alpha = self.alpha
        I = self.I
        n = self.n
        Ks = self.Ks
        m = 1.0 - 1.0 / n
        return Ks, alpha, I, n, m

    def __call__(self, u):
        Ks, alpha, I, n, m = self._get_params()

        P_p, P_n = _get_projections(u)  # Compute the positive/negative domains
        theta_e = 1.0 / ((1.0 + abs(alpha * u) ** n) ** m)
        f_p = P_p * np.ones(len(u)) * Ks  # ensures scalar Ks works
        f_n = P_n * Ks * theta_e ** I * ((1.0 - (1.0 - theta_e ** (1.0 / m)) ** m) ** 2)
        return f_p + f_n

    def derivM(self, u):
        """derivative with respect to m

        .. code::

            import sympy as sy

            alpha, u, n, I, Ks, theta_r, theta_s = sy.symbols(
                'alpha u n I Ks theta_r theta_s', real=True
            )

            m = 1.0 - 1.0/n
            theta_e = 1.0 / ((1.0 + sy.functions.Abs(alpha * u) ** n) ** m)

            f_n = Ks * theta_e ** I * (
                (1.0 - (1.0 - theta_e ** (1.0 / m)) ** m) ** 2
            )

            f_n = (
                (
                    theta_s - theta_r
                ) /
                (
                    (1.0 + abs(alpha * u)**n) ** (1.0 - 1.0 / n)
                ) +
                theta_r
            )
        """
        return (
            self._derivKs(u) + self._derivI(u) + self._derivN(u) + self._derivAlpha(u)
        )

    def _derivKs(self, u):
        if self.KsMap is None:
            return Utils.Zero()

        Ks, alpha, I, n, m = self._get_params()
        P_p, P_n = _get_projections(u)  # Compute the positive/negative domains
        theta_e = 1.0 / ((1.0 + abs(alpha * u) ** n) ** m)
        dKs_dm_p = P_p * self.KsDeriv
        dKs_dm_n = (
            P_n
            * Utils.sdiag(
                theta_e ** I * ((1.0 - (1.0 - theta_e ** (1.0 / m)) ** m) ** 2)
            )
            * self.KsDeriv
        )
        return dKs_dm_p + dKs_dm_n

    def _derivAlpha(self, u):
        if self.alphaMap is None:
            return Utils.Zero()
        Ks, alpha, I, n, m = self._get_params()
        ddm = I * u * n * Ks * abs(alpha * u) ** (n - 1) * np.sign(alpha * u) * (
            1.0 / n - 1
        ) * ((abs(alpha * u) ** n + 1) ** (1.0 / n - 1)) ** (I - 1) * (
            (
                1
                - 1.0
                / ((abs(alpha * u) ** n + 1) ** (1.0 / n - 1)) ** (1.0 / (1.0 / n - 1))
            )
            ** (1 - 1.0 / n)
            - 1
        ) ** 2 * (
            abs(alpha * u) ** n + 1
        ) ** (
            1.0 / n - 2
        ) - (
            2
            * u
            * n
            * Ks
            * abs(alpha * u) ** (n - 1)
            * np.sign(alpha * u)
            * (1.0 / n - 1)
            * ((abs(alpha * u) ** n + 1) ** (1.0 / n - 1)) ** I
            * (
                (
                    1
                    - 1.0
                    / ((abs(alpha * u) ** n + 1) ** (1.0 / n - 1))
                    ** (1.0 / (1.0 / n - 1))
                )
                ** (1 - 1.0 / n)
                - 1
            )
            * (abs(alpha * u) ** n + 1) ** (1.0 / n - 2)
        ) / (
            ((abs(alpha * u) ** n + 1) ** (1.0 / n - 1)) ** (1.0 / (1.0 / n - 1) + 1)
            * (
                1
                - 1.0
                / ((abs(alpha * u) ** n + 1) ** (1.0 / n - 1)) ** (1.0 / (1.0 / n - 1))
            )
            ** (1.0 / n)
        )
        ddm[u >= 0] = 0
        dA = Utils.sdiag(ddm) * self.alphaDeriv
        return dA

    def _derivN(self, u):
        if self.nMap is None:
            return Utils.Zero()
        Ks, alpha, I, n, m = self._get_params()
        ddm = 1.0 * I * Ks * (
            1.0 * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n)
        ) ** I * (
            (-1.0 + 1.0 / n)
            * np.log(abs(alpha * u))
            * abs(alpha * u) ** n
            / (abs(alpha * u) ** n + 1.0)
            - 1.0 * np.log(abs(alpha * u) ** n + 1.0) / n ** 2
        ) * (
            -(
                (
                    -(
                        (1.0 * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n))
                        ** (1.0 / (1.0 - 1.0 / n))
                    )
                    + 1.0
                )
                ** (1.0 - 1.0 / n)
            )
            + 1.0
        ) ** 2 * (
            abs(alpha * u) ** n + 1.0
        ) ** (
            -1.0 + 1.0 / n
        ) * (
            abs(alpha * u) ** n + 1.0
        ) ** (
            1.0 - 1.0 / n
        ) - 2 * Ks * (
            1.0 * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n)
        ) ** I * (
            -(
                (1.0 * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n))
                ** (1.0 / (1.0 - 1.0 / n))
            )
            + 1.0
        ) ** (
            1.0 - 1.0 / n
        ) * (
            -(
                (1.0 * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n))
                ** (1.0 / (1.0 - 1.0 / n))
            )
            * (1.0 - 1.0 / n)
            * (
                1.0
                * (
                    (-1.0 + 1.0 / n)
                    * np.log(abs(alpha * u))
                    * abs(alpha * u) ** n
                    / (abs(alpha * u) ** n + 1.0)
                    - 1.0 * np.log(abs(alpha * u) ** n + 1.0) / n ** 2
                )
                * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n)
                * (abs(alpha * u) ** n + 1.0) ** (1.0 - 1.0 / n)
                / (1.0 - 1.0 / n)
                - 1.0
                * np.log(1.0 * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n))
                / (n ** 2 * (1.0 - 1.0 / n) ** 2)
            )
            / (
                -(
                    (1.0 * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n))
                    ** (1.0 / (1.0 - 1.0 / n))
                )
                + 1.0
            )
            + 1.0
            * np.log(
                -(
                    (1.0 * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n))
                    ** (1.0 / (1.0 - 1.0 / n))
                )
                + 1.0
            )
            / n ** 2
        ) * (
            -(
                (
                    -(
                        (1.0 * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n))
                        ** (1.0 / (1.0 - 1.0 / n))
                    )
                    + 1.0
                )
                ** (1.0 - 1.0 / n)
            )
            + 1.0
        )
        ddm[u >= 0] = 0
        dn = Utils.sdiag(ddm) * self.nDeriv
        return dn

    def _derivI(self, u):
        if self.IMap is None:
            return Utils.Zero()
        Ks, alpha, I, n, m = self._get_params()
        ddm = (
            Ks
            * (1.0 * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n)) ** I
            * (
                -(
                    (
                        -(
                            (1.0 * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n))
                            ** (1.0 / (1.0 - 1.0 / n))
                        )
                        + 1.0
                    )
                    ** (1.0 - 1.0 / n)
                )
                + 1.0
            )
            ** 2
            * np.log(1.0 * (abs(alpha * u) ** n + 1.0) ** (-1.0 + 1.0 / n))
        )
        ddm[u >= 0] = 0
        dI = Utils.sdiag(ddm) * self.IDeriv
        return dI

    def derivU(self, u):
        Ks, alpha, I, n, m = self._get_params()
        ddm = I * alpha * n * Ks * abs(alpha * u) ** (n - 1.0) * np.sign(alpha * u) * (
            1.0 / n - 1.0
        ) * ((abs(alpha * u) ** n + 1) ** (1.0 / n - 1)) ** (I - 1) * (
            (
                1
                - 1.0
                / ((abs(alpha * u) ** n + 1) ** (1.0 / n - 1)) ** (1.0 / (1.0 / n - 1))
            )
            ** (1 - 1.0 / n)
            - 1
        ) ** 2 * (
            abs(alpha * u) ** n + 1
        ) ** (
            1.0 / n - 2
        ) - (
            2
            * alpha
            * n
            * Ks
            * abs(alpha * u) ** (n - 1)
            * np.sign(alpha * u)
            * (1.0 / n - 1)
            * ((abs(alpha * u) ** n + 1) ** (1.0 / n - 1)) ** I
            * (
                (
                    1
                    - 1.0
                    / ((abs(alpha * u) ** n + 1) ** (1.0 / n - 1))
                    ** (1.0 / (1.0 / n - 1))
                )
                ** (1 - 1.0 / n)
                - 1
            )
            * (abs(alpha * u) ** n + 1) ** (1.0 / n - 2)
        ) / (
            ((abs(alpha * u) ** n + 1) ** (1.0 / n - 1)) ** (1.0 / (1.0 / n - 1) + 1)
            * (
                1
                - 1.0
                / ((abs(alpha * u) ** n + 1) ** (1.0 / n - 1)) ** (1.0 / (1.0 / n - 1))
            )
            ** (1.0 / n)
        )
        ddm[u >= 0] = 0
        g = Utils.sdiag(ddm)
        return g


def van_genuchten(mesh, **kwargs):
    return _partition_args(
        mesh,
        Vangenuchten_k,
        Vangenuchten_theta,
        ["alpha", "n", "Ks", "I"],
        ["alpha", "n", "theta_r", "theta_s"],
        **kwargs,
    )


class VanGenuchtenParams:
    """The RETC code for quantifying the hydraulic functions of unsaturated
    soils, Van Genuchten, M Th, Leij, F J, Yates, S R

    Table 3: Average values for selected soil water retention and hydraulic
    conductivity parameters for 11 major soil textural groups
    according to Rawls et al. [1982]
    """

    @property
    def sand(self):
        return {
            "theta_r": 0.020,
            "theta_s": 0.417,
            "alpha": 0.138 * 100.0,
            "n": 1.592,
            "Ks": 504.0 * constants.centi / constants.day,
        }

    @property
    def loamy_sand(self):
        return {
            "theta_r": 0.035,
            "theta_s": 0.401,
            "alpha": 0.115 * 100.0,
            "n": 1.474,
            "Ks": 146.6 * constants.centi / constants.day,
        }

    @property
    def sandy_loam(self):
        return {
            "theta_r": 0.041,
            "theta_s": 0.412,
            "alpha": 0.068 * 100.0,
            "n": 1.322,
            "Ks": 62.16 * constants.centi / constants.day,
        }

    @property
    def loam(self):
        return {
            "theta_r": 0.027,
            "theta_s": 0.434,
            "alpha": 0.090 * 100.0,
            "n": 1.220,
            "Ks": 16.32 * constants.centi / constants.day,
        }

    @property
    def silt_loam(self):
        return {
            "theta_r": 0.015,
            "theta_s": 0.486,
            "alpha": 0.048 * 100.0,
            "n": 1.211,
            "Ks": 31.68 * constants.centi / constants.day,
        }

    @property
    def sandy_clay_loam(self):
        return {
            "theta_r": 0.068,
            "theta_s": 0.330,
            "alpha": 0.036 * 100.0,
            "n": 1.250,
            "Ks": 10.32 * constants.centi / constants.day,
        }

    @property
    def clay_loam(self):
        return {
            "theta_r": 0.075,
            "theta_s": 0.390,
            "alpha": 0.039 * 100.0,
            "n": 1.194,
            "Ks": 5.52 * constants.centi / constants.day,
        }

    @property
    def silty_clay_loam(self):
        return {
            "theta_r": 0.040,
            "theta_s": 0.432,
            "alpha": 0.031 * 100.0,
            "n": 1.151,
            "Ks": 3.60 * constants.centi / constants.day,
        }

    @property
    def sandy_clay(self):
        return {
            "theta_r": 0.109,
            "theta_s": 0.321,
            "alpha": 0.034 * 100.0,
            "n": 1.168,
            "Ks": 2.88 * constants.centi / constants.day,
        }

    @property
    def silty_clay(self):
        return {
            "theta_r": 0.056,
            "theta_s": 0.423,
            "alpha": 0.029 * 100.0,
            "n": 1.127,
            "Ks": 2.16 * constants.centi / constants.day,
        }

    @property
    def clay(self):
        return {
            "theta_r": 0.090,
            "theta_s": 0.385,
            "alpha": 0.027 * 100.0,
            "n": 1.131,
            "Ks": 1.44 * constants.centi / constants.day,
        }

    # From:
    #   INDIRECT METHODS FOR ESTIMATING THE HYDRAULIC
    #   PROPERTIES OF UNSATURATED SOILS
    # @property
    # def siltLoamGE3(self):
    #     """Soil Index: 3310"""
    #     return {"theta_r": 0.139, "theta_s": 0.394, "alpha": 0.00414, "n": 2.15}
    # @property
    # def yoloLightClayK_WC(self):
    #     """Soil Index: None"""
    #     return {"theta_r": 0.205, "theta_s": 0.499, "alpha": 0.02793, "n": 1.71}
    # @property
    # def yoloLightClayK_H(self):
    #     """Soil Index: None"""
    #     return {"theta_r": 0.205, "theta_s": 0.499, "alpha": 0.02793, "n": 1.71}
    # @property
    # def hygieneSandstone(self):
    #     """Soil Index: 4130"""
    #     return {"theta_r": 0.000, "theta_s": 0.256, "alpha": 0.00562, "n": 3.27}
    # @property
    # def lambcrgClay(self):
    #     """Soil Index: 1003"""
    #     return {"theta_r": 0.000, "theta_s": 0.502, "alpha": 0.140, "n": 1.93}
    # @property
    # def beitNetofaClaySoil(self):
    #     """Soil Index: 1006"""
    #     return {"theta_r": 0.000, "theta_s": 0.447, "alpha": 0.00156, "n": 1.17}
    # @property
    # def shiohotSiltyClay(self):
    #     """Soil Index: 1101"""
    #     return {"theta_r": 0.000, "theta_s": 0.456, "alpha": 183, "n":1.17}
    # @property
    # def siltColumbia(self):
    #     """Soil Index: 2001"""
    #     return {"theta_r": 0.146, "theta_s": 0.397,  "alpha": 0.0145, "n": 1.85}
    # @property
    # def siltMontCenis(self):
    #     """Soil Index: 2002"""
    #     return {"theta_r": 0.000, "theta_s": 0.425, "alpha": 0.0103, "n": 1.34}
    # @property
    # def slateDust(self):
    #     """Soil Index: 2004"""
    #     return {"theta_r": 0.000, "theta_s": 0.498, "alpha": 0.00981, "n": 6.75}
    # @property
    # def weldSiltyClayLoam(self):
    #     """Soil Index: 3001"""
    #     return {"theta_r": 0.159, "theta_s": 0.496, "alpha": 0.0136, "n": 5.45}
    # @property
    # def rideauClayLoam_Wetting(self):
    #     """Soil Index: 3101a"""
    #     return {"theta_r": 0.279, "theta_s": 0.419, "alpha": 0.0661, "n": 1.89}
    # @property
    # def rideauClayLoam_Drying(self):
    #     """Soil Index: 3101b"""
    #     return {"theta_r": 0.290, "theta_s": 0.419, "alpha": 0.0177, "n": 3.18}
    # @property
    # def caribouSiltLoam_Drying(self):
    #     """Soil Index: 3301a"""
    #     return {"theta_r": 0.000, "theta_s": 0.451, "alpha": 0.00845, "n": 1.29}
    # @property
    # def caribouSiltLoam_Wetting(self):
    #     """Soil Index: 3301b"""
    #     return {"theta_r": 0.000, "theta_s": 0.450, "alpha": 0.140, "n": 1.09}
    # @property
    # def grenvilleSiltLoam_Wetting(self):
    #     """Soil Index: 3302a"""
    #     return {"theta_r": 0.013, "theta_s": 0523,  "alpha": 0.0630, "n": 1.24}
    # @property
    # def grenvilleSiltLoam_Drying(self):
    #     """Soil Index: 3302c"""
    #     return {"theta_r": 0.000, "theta_s": 0.488, "alpha": 0.0112, "n": 1.23}
    # @property
    # def touchetSiltLoam(self):
    #     """Soil Index: 3304"""
    #     return {"theta_r": 0.183, "theta_s": 0.498, "alpha": 0.0104, "n": 5.78}
    # @property
    # def gilatLoam(self):
    #     """Soil Index: 3402a"""
    #     return {"theta_r": 0.000, "theta_s": 0.454, "alpha": 0.0291, "n": 1.47}
    # @property
    # def pachapaLoam(self):
    #     """Soil Index: 3403"""
    #     return {"theta_r": 0.000, "theta_s": 0.472, "alpha": 0.00829, "n": 1.62}
    # @property
    # def adelantoLoam(self):
    #     """Soil Index: 3404"""
    #     return {"theta_r": 0.000, "theta_s": 0.444, "alpha": 0.00710, "n": 1.26}
    # @property
    # def indioLoam(self):
    #     """Soil Index: 3405a"""
    #     return {"theta_r": 0.000, "theta_s": 0.507, "alpha": 0.00847, "n": 1.60}
    # @property
    # def guclphLoam(self):
    #     """Soil Index: 3407a"""
    #     return {"theta_r": 0.000, "theta_s": 0.563, "alpha": 0.0275, "n": 1.27}
    # @property
    # def guclphLoam(self):
    #     """Soil Index: 3407b"""
    #     return {"theta_r": 0.236, "theta_s": 0.435, "alpha": 0.0271, "n": 262}
    # @property
    # def rubiconSandyLoam(self):
    #     """Soil Index: 3501a"""
    #     return {"theta_r": 0.000, "theta_s": 0.393,  "alpha": 0.00972, "n": 2.18}
    # @property
    # def rubiconSandyLoam(self):
    #     """Soil Index: 350lb"""
    #     return {"theta_r": 0.000, "theta_s": 0.433, "alpha": 0.147, "n": 1.28}
    # @property
    # def pachapaFmeSandyClay(self):
    #     """Soil Index: 3503a"""
    #     return {"theta_r": 0.000, "theta_s": 0.340, "alpha": 0.0194, "n": 1.45}
    # @property
    # def gilatSandyLoam(self):
    #     """Soil Index: 3504"""
    #     return {"theta_r": 0.000, "theta_s": 0.432, "alpha": 0.0103, "n": 1.48}
    # @property
    # def plainfieldSand_210to250(self):
    #     """Soil Index: 4101a"""
    #     return {"theta_r": 0.000, "theta_s": 0.351, "alpha": 0.0236, "n": 12.30}
    # @property
    # def plainfieldSand_210to250(self):
    #     """Soil Index: 4101b"""
    #     return {"theta_r": 0.000, "theta_s": 0.312, "alpha": 0.0387, "n": 4.48}
    # @property
    # def plainfieldSand_177to210(self):
    #     """Soil Index: 4102a"""
    #     return {"theta_r": 0.000, "theta_s": 0.361, "alpha": 0.0207, "n": 10.0}
    # @property
    # def plainfieldSand_177to210(self):
    #     """Soil Index: 4102b"""
    #     return {"theta_r": 0.022, "theta_s": 0.309, "alpha": 0.0328, "n": 6.23}
    # @property
    # def plainfieldSand_149to177(self):
    #     """Soil Index: 4103a"""
    #     return {"theta_r": 0.000, "theta_s": 0.387, "alpha": 0.0173, "n": 7.80}
    # @property
    # def plainfieldSand_149to177(self):
    #     """Soil Index: 4103b"""
    #     return {"theta_r": 0.025, "theta_s": 0.321, "alpha": 0.0272, "n": 6.69}
    # @property
    # def plainfieldSand_l25to149(self):
    #     """Soil Index: 4104a"""
    #     return {"theta_r": 0.000, "theta_s": 03770, "alpha": 0.0145, "n": 10.60}
    # @property
    # def plainfieldSand_125to149(self):
    #     """Soil Index: 4104b"""
    #     return {"theta_r": 0.000, "theta_s": 0.342, "alpha": 0.0230, "n": 5.18}



import numpy as np
import properties

from . import Utils
from . import Survey
from . import ObjectiveFunction
from .Tests import checkDerivative
from scipy.sparse import csr_matrix as csr
import dask
import dask.array as da


class BaseDataMisfit(ObjectiveFunction.L2ObjectiveFunction):
    """
    BaseDataMisfit

        .. note::

            You should inherit from this class to create your own data misfit
            term.
    """

    debug = False  #: Print debugging information
    counter = None  #: Set this to a SimPEG.Utils.Counter() if you want to count things

    _hasFields = True  #: Data Misfits take fields, handy to store them

    def __init__(self, survey, **kwargs):
        assert survey.ispaired, "The survey must be paired to a problem."
        if isinstance(survey, Survey.BaseSurvey):
            self.survey = survey
            self.prob = survey.prob
        super().__init__(**kwargs)

    @property
    def nP(self):
        if self._mapping is not None:
            return self.mapping.nP
        elif self.prob.model is not None:
            return len(self.prob.model)
        else:
            return "*"

    @property
    def Wd(self):
        raise AttributeError(
            "The `Wd` property been depreciated, please use: `W` instead"
        )


class l2_DataMisfit(BaseDataMisfit):
    """

    The data misfit with an l_2 norm:

    .. math::

        \\mu_\\text{data} = {1\\over 2}\\left|
        \\mathbf{W}_d (\\mathbf{d}_\\text{pred} -
        \\mathbf{d}_\\text{obs}) \\right|_2^2

    """

    std = 0.05  #: default standard deviation if not provided by survey
    eps = None  #: default floor
    eps_factor = 1e-5  #: factor to multiply by the norm of the data to create floor
    scale = 1.0

    def __init__(self, survey, **kwargs):
        BaseDataMisfit.__init__(self, survey, **kwargs)

        if self.std is None:
            if getattr(self.survey, "std", None) is None:
                print("SimPEG.DataMisfit.l2_DataMisfit assigning default std " "of 5%")
            else:
                self.std = self.survey.std

        if self.eps is None:
            if getattr(self.survey, "eps", None) is None:
                self.eps = (
                    np.linalg.norm(Utils.mkvc(survey.dobs), 2) * self.eps_factor
                )  # default
            else:
                self.eps = self.survey.eps

    @property
    def W(self):
        """W

            The data weighting matrix.

            The default is based on the norm of the data plus a noise floor.

            :rtype: scipy.sparse.csr_matrix
            :return: W

        """

        if getattr(self, "_W", None) is None:

            survey = self.survey
            self._W = Utils.sdiag(1 / (abs(survey.dobs) * self.std + self.eps))

        return self._W

    @W.setter
    def W(self, value):
        if len(value.shape) < 2:
            value = Utils.sdiag(value)
        assert value.shape == (
            self.survey.nD,
            self.survey.nD,
        ), "W must have shape ({nD},{nD}), not ({val0}, val{1})".format(
            nD=self.survey.nD, val0=value.shape[0], val1=value.shape[1]
        )
        self._W = value

    @Utils.timeIt
    def __call__(self, m, f=None):
        "__call__(m, f=None)"
        if f is None:
            f = self.prob.fields(m)

        W_res = dask.delayed(csr.dot)(self.W, self.survey.residual(m, f))
        vec = da.from_delayed(W_res, dtype=float, shape=[self.W.shape[1]])
        #        R = self.W * self.survey.residual(m, f)
        return 0.5 * da.dot(vec, vec)

    @Utils.timeIt
    def deriv(self, m, f=None):
        """
        deriv(m, f=None)

        Derivative of the data misfit

        .. math::

            \\mathbf{J}^{\top} \\mathbf{W}^{\top} \\mathbf{W}
            (\\mathbf{d} - \\mathbf{d}^{obs})

        :param numpy.ndarray m: model
        :param SimPEG.Fields.Fields f: fields object
        """
        if f is None:
            f = self.prob.fields(m)

        w_d = dask.delayed(csr.dot)(self.W, self.survey.residual(m, f=f))

        wtw_d = dask.delayed(csr.dot)(self.scale * w_d, self.W)

        row = da.from_delayed(wtw_d, dtype=float, shape=[self.W.shape[0]])
        return self.prob.Jtvec(m, row, f=f)

    @Utils.timeIt
    def deriv2(self, m, v, f=None):
        """
        deriv2(m, v, f=None)

        .. math::

            \\mathbf{J}^{\top} \\mathbf{W}^{\top} \\mathbf{W} \\mathbf{J}

        :param numpy.ndarray m: model
        :param numpy.ndarray v: vector
        :param SimPEG.Fields.Fields f: fields object
        """
        if f is None:
            f = self.prob.fields(m)

        jtvec = self.prob.Jvec_approx(m, v, f=f)

        w_jtvec = dask.delayed(csr.dot)(self.W, jtvec)

        wtw_jtvec = dask.delayed(csr.dot)(self.scale * w_jtvec, self.W)

        row = da.from_delayed(wtw_jtvec, dtype=float, shape=[self.W.shape[0]])
        return self.prob.Jtvec_approx(m, row, f=f)


# class l1_DataMisfit(l2_DataMisfit):
#     """

#     The data misfit with an l_2 norm:

#     .. math::

#         \mu_\\text{data} = {1\over 2}\left|
#         \mathbf{W}_d (\mathbf{d}_\\text{pred} -
#         \mathbf{d}_\\text{obs}) \\right|_2^2

#     """
#     def __init__(self, survey, **kwargs):
#         self._stashedR = None
#         self.epsilon = 1e-8
#         self.norm = 1.
#         l2_DataMisfit.__init__(self, survey, **kwargs)

#     @property
#     def stashedR(self):
#         return self._stashedR

#     @stashedR.setter
#     def stashedR(self, value):
#         self._stashedR = value

#     def R(self, f_m):
#         # if R is stashed, return that instead
#         if getattr(self, 'stashedR') is not None:
#             return self.stashedR

#         # Eta scaling is important for mix-norms...do not mess with it
#         eta = (2. * np.abs(f_m).max() * self.epsilon)**(1.-self.norm/2.)
#         r = (eta / (f_m**2. + self.epsilon**2.)**(1.-(self.norm)/2.))**0.5
#         self.stashedR = r  # stash on the first calculation
#         return r

#     @Utils.timeIt
#     def __call__(self, m, f=None):
#         "__call__(m, f=None)"
#         if f is None:
#             f = self.prob.fields(m)
#         R = self.W * self.survey.residual(m, f)

#         r = self.R(f)
#         R = R*Utils.sdiag(r)

#         return 0.5*np.vdot(R, R)

#     @Utils.timeIt
#     def deriv(self, m, f=None):
#         """
#         deriv(m, f=None)

#         Derivative of the data misfit

#         .. math::

#             \mathbf{J}^{\top} \mathbf{W}^{\top} \mathbf{W}
#             (\mathbf{d} - \mathbf{d}^{obs})

#         :param numpy.ndarray m: model
#         :param SimPEG.Fields.Fields f: fields object
#         """
#         if f is None:
#             f = self.prob.fields(m)

#         r = self.R(f)

#         R = Utils.sdiag(r)

#         return self.prob.Jtvec(
#             m, self.W.T * R.T * (self.scale * R * self.W * self.survey.residual(m, f=f)), f=f
#         )

#     @Utils.timeIt
#     def deriv2(self, m, v, f=None):
#         """
#         deriv2(m, v, f=None)

#         .. math::

#             \mathbf{J}^{\top} \mathbf{W}^{\top} \mathbf{W} \mathbf{J}

#         :param numpy.ndarray m: model
#         :param numpy.ndarray v: vector
#         :param SimPEG.Fields.Fields f: fields object
#         """
#         # if f is None:
#         f = self.prob.fields(m)

#         r = self.R(f)

#         R = Utils.sdiag(r)

#         return self.prob.Jtvec_approx(
#             m, self.W * R.T * (self.scale * R * self.W * self.prob.Jvec_approx(m, v, f=f)), f=f
#         )



""" Module RxNSEM.py

Receivers for the NSEM problem

"""

from scipy.constants import mu_0

import SimPEG
import numpy as np
from .. import mkvc


class BaseRxNSEM_Point(SimPEG.Survey.BaseRx):
    """
    Natural source receiver base class.

    Assumes that the data locations are xyz coordinates.

    :param numpy.ndarray locs: receiver locations (ie. :code:`np.r_[x,y,z]`)
    :param string orientation: receiver orientation 'x', 'y' or 'z'
    :param string component: real or imaginary component 'real' or 'imag'
    """

    def __init__(self, locs, orientation=None, component=None):
        assert orientation in [
            "xx",
            "xy",
            "yx",
            "yy",
            "zx",
            "zy",
        ], "Orientation {!s} not known. Orientation must be in 'x', 'y', 'z'. Arbitrary orientations have not yet been implemented.".format(
            orientation
        )
        assert component in [
            "real",
            "imag",
        ], f"'component' must be 'real' or 'imag', not {component!s}"

        self.orientation = orientation
        self.component = component

        SimPEG.Survey.BaseRx.__init__(
            self, locs, rxType=None
        )  # TODO: remove rxType from baseRx

    # Set a mesh property
    @property
    def mesh(self):
        return self._mesh

    @mesh.setter
    def mesh(self, value):
        if value is getattr(self, "_mesh", None):
            pass
        else:
            self._mesh = value

    @property
    def src(self):
        return self._src

    @src.setter
    def src(self, value):
        self._src = value

    @property
    def f(self):
        return self._f

    @f.setter
    def f(self, value):
        self._f = value

    def _locs_e(self):
        if self.locs.ndim == 3:
            loc = self.locs[:, :, 0]
        else:
            loc = self.locs
        return loc

    def _locs_b(self):
        if self.locs.ndim == 3:
            loc = self.locs[:, :, 1]
        else:
            loc = self.locs
        return loc

    # Location projection
    @property
    def Pex(self):
        if getattr(self, "_Pex", None) is None:
            self._Pex = self._mesh.getInterpolationMat(self._locs_e(), "Ex")
        return self._Pex

    @property
    def Pey(self):
        if getattr(self, "_Pey", None) is None:
            self._Pey = self._mesh.getInterpolationMat(self._locs_e(), "Ey")
        return self._Pey

    @property
    def Pbx(self):
        if getattr(self, "_Pbx", None) is None:
            self._Pbx = self._mesh.getInterpolationMat(self._locs_b(), "Fx")
        return self._Pbx

    @property
    def Pby(self):
        if getattr(self, "_Pby", None) is None:
            self._Pby = self._mesh.getInterpolationMat(self._locs_b(), "Fy")
        return self._Pby

    @property
    def Pbz(self):
        if getattr(self, "_Pbz", None) is None:
            self._Pbz = self._mesh.getInterpolationMat(self._locs_e(), "Fz")
        return self._Pbz

    # Utility for convienece
    def _sDiag(self, t):
        return SimPEG.Utils.sdiag(mkvc(t, 2))

    # Get the components of the fields
    # px: x-polaration and py: y-polaration.
    @property
    def _ex_px(self):
        return self.Pex * self.f[self.src, "e_px"]

    @property
    def _ey_px(self):
        return self.Pey * self.f[self.src, "e_px"]

    @property
    def _ex_py(self):
        return self.Pex * self.f[self.src, "e_py"]

    @property
    def _ey_py(self):
        return self.Pey * self.f[self.src, "e_py"]

    @property
    def _hx_px(self):
        return self.Pbx * self.f[self.src, "b_px"] / mu_0

    @property
    def _hy_px(self):
        return self.Pby * self.f[self.src, "b_px"] / mu_0

    @property
    def _hz_px(self):
        return self.Pbz * self.f[self.src, "b_px"] / mu_0

    @property
    def _hx_py(self):
        return self.Pbx * self.f[self.src, "b_py"] / mu_0

    @property
    def _hy_py(self):
        return self.Pby * self.f[self.src, "b_py"] / mu_0

    @property
    def _hz_py(self):
        return self.Pbz * self.f[self.src, "b_py"] / mu_0

    # Get the derivatives

    def _ex_px_u(self, vec):
        return self.Pex * self.f._e_pxDeriv_u(self.src, vec)

    def _ey_px_u(self, vec):
        return self.Pey * self.f._e_pxDeriv_u(self.src, vec)

    def _ex_py_u(self, vec):
        return self.Pex * self.f._e_pyDeriv_u(self.src, vec)

    def _ey_py_u(self, vec):
        return self.Pey * self.f._e_pyDeriv_u(self.src, vec)

    def _hx_px_u(self, vec):
        return self.Pbx * self.f._b_pxDeriv_u(self.src, vec) / mu_0

    def _hy_px_u(self, vec):
        return self.Pby * self.f._b_pxDeriv_u(self.src, vec) / mu_0

    def _hz_px_u(self, vec):
        return self.Pbz * self.f._b_pxDeriv_u(self.src, vec) / mu_0

    def _hx_py_u(self, vec):
        return self.Pbx * self.f._b_pyDeriv_u(self.src, vec) / mu_0

    def _hy_py_u(self, vec):
        return self.Pby * self.f._b_pyDeriv_u(self.src, vec) / mu_0

    def _hz_py_u(self, vec):
        return self.Pbz * self.f._b_pyDeriv_u(self.src, vec) / mu_0

    # Define the components of the derivative

    @property
    def _Hd(self):
        return self._sDiag(
            1.0
            / (
                self._sDiag(self._hx_px) * self._hy_py
                - self._sDiag(self._hx_py) * self._hy_px
            )
        )

    def _Hd_uV(self, v):
        return (
            self._sDiag(self._hy_py) * self._hx_px_u(v)
            + self._sDiag(self._hx_px) * self._hy_py_u(v)
            - self._sDiag(self._hx_py) * self._hy_px_u(v)
            - self._sDiag(self._hy_px) * self._hx_py_u(v)
        )

    # Adjoint
    @property
    def _aex_px(self):
        return mkvc(mkvc(self.f[self.src, "e_px"], 2).T * self.Pex.T)

    @property
    def _aey_px(self):
        return mkvc(mkvc(self.f[self.src, "e_px"], 2).T * self.Pey.T)

    @property
    def _aex_py(self):
        return mkvc(mkvc(self.f[self.src, "e_py"], 2).T * self.Pex.T)

    @property
    def _aey_py(self):
        return mkvc(mkvc(self.f[self.src, "e_py"], 2).T * self.Pey.T)

    @property
    def _ahx_px(self):
        return mkvc(mkvc(self.f[self.src, "b_px"], 2).T / mu_0 * self.Pbx.T)

    @property
    def _ahy_px(self):
        return mkvc(mkvc(self.f[self.src, "b_px"], 2).T / mu_0 * self.Pby.T)

    @property
    def _ahz_px(self):
        return mkvc(mkvc(self.f[self.src, "b_px"], 2).T / mu_0 * self.Pbz.T)

    @property
    def _ahx_py(self):
        return mkvc(mkvc(self.f[self.src, "b_py"], 2).T / mu_0 * self.Pbx.T)

    @property
    def _ahy_py(self):
        return mkvc(mkvc(self.f[self.src, "b_py"], 2).T / mu_0 * self.Pby.T)

    @property
    def _ahz_py(self):
        return mkvc(mkvc(self.f[self.src, "b_py"], 2).T / mu_0 * self.Pbz.T)

    # NOTE: need to add a .T at the end for the output to be (nU,)
    def _aex_px_u(self, vec):
        """
        """
        # vec is (nD,) and returns a (nU,)
        return self.f._e_pxDeriv_u(self.src, self.Pex.T * mkvc(vec,), adjoint=True)

    def _aey_px_u(self, vec):
        """
        """
        # vec is (nD,) and returns a (nU,)
        return self.f._e_pxDeriv_u(self.src, self.Pey.T * mkvc(vec,), adjoint=True)

    def _aex_py_u(self, vec):
        """
        """
        # vec is (nD,) and returns a (nU,)
        return self.f._e_pyDeriv_u(self.src, self.Pex.T * mkvc(vec,), adjoint=True)

    def _aey_py_u(self, vec):
        """
        """
        # vec is (nD,) and returns a (nU,)
        return self.f._e_pyDeriv_u(self.src, self.Pey.T * mkvc(vec,), adjoint=True)

    def _ahx_px_u(self, vec):
        """
        """
        # vec is (nD,) and returns a (nU,)
        return (
            self.f._b_pxDeriv_u(self.src, self.Pbx.T * mkvc(vec,), adjoint=True) / mu_0
        )

    def _ahy_px_u(self, vec):
        """
        """
        # vec is (nD,) and returns a (nU,)
        return (
            self.f._b_pxDeriv_u(self.src, self.Pby.T * mkvc(vec,), adjoint=True) / mu_0
        )

    def _ahz_px_u(self, vec):
        """
        """
        # vec is (nD,) and returns a (nU,)
        return (
            self.f._b_pxDeriv_u(self.src, self.Pbz.T * mkvc(vec,), adjoint=True) / mu_0
        )

    def _ahx_py_u(self, vec):
        """
        """
        # vec is (nD,) and returns a (nU,)
        return (
            self.f._b_pyDeriv_u(self.src, self.Pbx.T * mkvc(vec,), adjoint=True) / mu_0
        )

    def _ahy_py_u(self, vec):
        """
        """
        # vec is (nD,) and returns a (nU,)
        return (
            self.f._b_pyDeriv_u(self.src, self.Pby.T * mkvc(vec,), adjoint=True) / mu_0
        )

    def _ahz_py_u(self, vec):
        """
        """
        # vec is (nD,) and returns a (nU,)
        return (
            self.f._b_pyDeriv_u(self.src, self.Pbz.T * mkvc(vec,), adjoint=True) / mu_0
        )

    # Define the components of the derivative
    @property
    def _aHd(self):
        return self._sDiag(
            1.0
            / (
                self._sDiag(self._ahx_px) * self._ahy_py
                - self._sDiag(self._ahx_py) * self._ahy_px
            )
        )

    def _aHd_uV(self, x):
        return (
            self._ahx_px_u(self._sDiag(self._ahy_py) * x)
            + self._ahx_px_u(self._sDiag(self._ahy_py) * x)
            - self._ahy_px_u(self._sDiag(self._ahx_py) * x)
            - self._ahx_py_u(self._sDiag(self._ahy_px) * x)
        )

    def eval(self, src, mesh, f, return_complex=False):
        """
        Function to evaluate datum for this receiver
        """
        raise NotImplementedError("SimPEG.EM.NSEM receiver has to have an eval method")

    def evalDeriv(self, src, mesh, f, v, adjoint=False):
        """
        Function to evaluate datum for this receiver
        """
        raise NotImplementedError(
            "SimPEG.EM.NSEM receiver has to have an evalDeriv method"
        )


class Point_impedance1D(SimPEG.Survey.BaseRx):
    """
    Natural source 1D impedance receiver class

    :param string component: real or imaginary component 'real' or 'imag'
    """

    orientation = "yx"

    def __init__(self, locs, component=None):
        assert component in [
            "real",
            "imag",
        ], f"'component' must be 'real' or 'imag', not {component!s}"

        self.component = component
        SimPEG.Survey.BaseRx.__init__(self, locs, rxType=None)

    @property
    def mesh(self):
        return self._mesh

    @mesh.setter
    def mesh(self, value):
        if value is getattr(self, "_mesh", None):
            pass
        else:
            self._mesh = value

    # Utility for convienece
    def _sDiag(self, t):
        return SimPEG.Utils.sdiag(mkvc(t, 2))

    @property
    def src(self):
        return self._src

    @src.setter
    def src(self, value):
        self._src = value

    @property
    def f(self):
        return self._f

    @f.setter
    def f(self, value):
        self._f = value

    @property
    def Pex(self):
        if getattr(self, "_Pex", None) is None:
            self._Pex = self._mesh.getInterpolationMat(self.locs[:, -1], "Fx")
        return self._Pex

    @property
    def Pbx(self):
        if getattr(self, "_Pbx", None) is None:
            self._Pbx = self._mesh.getInterpolationMat(self.locs[:, -1], "Ex")
        return self._Pbx

    @property
    def _ex(self):
        return self.Pex * mkvc(self.f[self.src, "e_1d"], 2)

    @property
    def _hx(self):
        return self.Pbx * mkvc(self.f[self.src, "b_1d"], 2) / mu_0

    def _ex_u(self, v):
        return self.Pex * self.f._eDeriv_u(self.src, v)

    def _hx_u(self, v):
        return self.Pbx * self.f._bDeriv_u(self.src, v) / mu_0

    def _aex_u(self, v):
        return self.f._eDeriv_u(self.src, self.Pex.T * v, adjoint=True)

    def _ahx_u(self, v):
        return self.f._bDeriv_u(self.src, self.Pbx.T * v, adjoint=True) / mu_0

    @property
    def _Hd(self):
        return self._sDiag(1.0 / self._hx)

    def eval(self, src, mesh, f, return_complex=False):
        """
        Project the fields to natural source data.

        :param SimPEG.EM.NSEM.SrcNSEM src: NSEM source
        :param discretize.TensorMesh mesh: Mesh defining the topology of the problem
        :param SimPEG.EM.NSEM.FieldsNSEM f: NSEM fields object of the source
        :param bool (optional) return_complex: Flag for return the complex evaluation
        :rtype: numpy.array
        :return: Evaluated data for the receiver
        """
        # NOTE: Maybe set this as a property
        self.src = src
        self.mesh = mesh
        self.f = f

        rx_eval_complex = -self._Hd * self._ex
        # Return the full impedance
        if return_complex:
            return rx_eval_complex
        return getattr(rx_eval_complex, self.component)

    def evalDeriv(self, src, mesh, f, v, adjoint=False):
        """method evalDeriv

        The derivative of the projection wrt u

        :param SimPEG.EM.NSEM.SrcNSEM src: NSEM source
        :param discretize.TensorMesh mesh: Mesh defining the topology of the problem
        :param SimPEG.EM.NSEM.FieldsNSEM f: NSEM fields object of the source
        :param numpy.ndarray v: vector of size (nU,) (adjoint=False)
            and size (nD,) (adjoint=True)
        :rtype: numpy.array
        :return: Calculated derivative (nD,) (adjoint=False) and (nP,2) (adjoint=True)
            for both polarizations
        """
        self.src = src
        self.mesh = mesh
        self.f = f

        if adjoint:
            Z1d = self.eval(src, mesh, f, True)

            def aZ_N_uV(x):
                return -self._aex_u(x)

            def aZ_D_uV(x):
                return self._ahx_u(x)

            rx_deriv = aZ_N_uV(self._Hd.T * v) - aZ_D_uV(
                self._sDiag(Z1d).T * self._Hd.T * v
            )
            if self.component == "imag":
                rx_deriv_component = 1j * rx_deriv
            elif self.component == "real":
                rx_deriv_component = rx_deriv.astype(complex)
        else:
            Z1d = self.eval(src, mesh, f, True)
            Z_N_uV = -self._ex_u(v)
            Z_D_uV = self._hx_u(v)
            # Evaluate
            rx_deriv = self._Hd * (Z_N_uV - self._sDiag(Z1d) * Z_D_uV)
            rx_deriv_component = np.array(getattr(rx_deriv, self.component))
        return rx_deriv_component


class Point_impedance3D(BaseRxNSEM_Point):
    """
    Natural source 3D impedance receiver class

    :param numpy.ndarray locs: receiver locations (ie. :code:`np.r_[x,y,z]`)
    :param string orientation: receiver orientation 'x', 'y' or 'z'
    :param string component: real or imaginary component 'real' or 'imag'
    """

    def __init__(self, locs, orientation=None, component=None):

        BaseRxNSEM_Point.__init__(
            self, locs, orientation=orientation, component=component
        )

    def eval(self, src, mesh, f, return_complex=False):
        """
        Project the fields to natural source data.

            :param SrcNSEM src: The source of the fields to project
            :param discretize.TensorMesh mesh: topological mesh corresponding to the fields
            :param FieldsNSEM f: Natural source fields object to project
            :rtype: numpy.array
            :return: component of the impedance evaluation
        """
        # NOTE: Maybe set this as a property
        self.src = src
        self.mesh = mesh
        self.f = f

        if "xx" in self.orientation:
            Zij = self._ex_px * self._hy_py - self._ex_py * self._hy_px
        elif "xy" in self.orientation:
            Zij = -self._ex_px * self._hx_py + self._ex_py * self._hx_px
        elif "yx" in self.orientation:
            Zij = self._ey_px * self._hy_py - self._ey_py * self._hy_px
        elif "yy" in self.orientation:
            Zij = -self._ey_px * self._hx_py + self._ey_py * self._hx_px
        # Calculate the complex value
        rx_eval_complex = self._Hd * Zij

        # Return the full impedance
        if return_complex:
            return rx_eval_complex
        return getattr(rx_eval_complex, self.component)

    def evalDeriv(self, src, mesh, f, v, adjoint=False):
        """
        The derivative of the projection wrt u

        :param SimPEG.EM.NSEM.SrcNSEM src: NSEM source
        :param discretize.TensorMesh mesh: Mesh defining the topology of the problem
        :param SimPEG.EM.NSEM.FieldsNSEM f: NSEM fields object of the source
        :param numpy.ndarray v: vector of size (nU,) (adjoint=False)
            and size (nD,) (adjoint=True)
        :rtype: numpy.array
        :return: Calculated derivative (nD,) (adjoint=False) and (nP,2) (adjoint=True)
            for both polarizations
        """
        self.src = src
        self.mesh = mesh
        self.f = f

        if adjoint:
            if "xx" in self.orientation:
                Zij = self._sDiag(
                    self._aHd
                    * (
                        self._sDiag(self._ahy_py) * self._aex_px
                        - self._sDiag(self._ahy_px) * self._aex_py
                    )
                )

                def ZijN_uV(x):
                    return (
                        self._aex_px_u(self._sDiag(self._ahy_py) * x)
                        + self._ahy_py_u(self._sDiag(self._aex_px) * x)
                        - self._ahy_px_u(self._sDiag(self._aex_py) * x)
                        - self._aex_py_u(self._sDiag(self._ahy_px) * x)
                    )

            elif "xy" in self.orientation:
                Zij = self._sDiag(
                    self._aHd
                    * (
                        -self._sDiag(self._ahx_py) * self._aex_px
                        + self._sDiag(self._ahx_px) * self._aex_py
                    )
                )

                def ZijN_uV(x):
                    return (
                        -self._aex_px_u(self._sDiag(self._ahx_py) * x)
                        - self._ahx_py_u(self._sDiag(self._aex_px) * x)
                        + self._ahx_px_u(self._sDiag(self._aex_py) * x)
                        + self._aex_py_u(self._sDiag(self._ahx_px) * x)
                    )

            elif "yx" in self.orientation:
                Zij = self._sDiag(
                    self._aHd
                    * (
                        self._sDiag(self._ahy_py) * self._aey_px
                        - self._sDiag(self._ahy_px) * self._aey_py
                    )
                )

                def ZijN_uV(x):
                    return (
                        self._aey_px_u(self._sDiag(self._ahy_py) * x)
                        + self._ahy_py_u(self._sDiag(self._aey_px) * x)
                        - self._ahy_px_u(self._sDiag(self._aey_py) * x)
                        - self._aey_py_u(self._sDiag(self._ahy_px) * x)
                    )

            elif "yy" in self.orientation:
                Zij = self._sDiag(
                    self._aHd
                    * (
                        -self._sDiag(self._ahx_py) * self._aey_px
                        + self._sDiag(self._ahx_px) * self._aey_py
                    )
                )

                def ZijN_uV(x):
                    return (
                        -self._aey_px_u(self._sDiag(self._ahx_py) * x)
                        - self._ahx_py_u(self._sDiag(self._aey_px) * x)
                        + self._ahx_px_u(self._sDiag(self._aey_py) * x)
                        + self._aey_py_u(self._sDiag(self._ahx_px) * x)
                    )

            # Calculate the complex derivative
            rx_deriv_real = ZijN_uV(self._aHd * v) - self._aHd_uV(Zij.T * self._aHd * v)
            # NOTE: Need to reshape the output to go from 2*nU array to a (nU,2) matrix for each polarization
            # rx_deriv_real = np.hstack((mkvc(rx_deriv_real[:len(rx_deriv_real)/2],2),mkvc(rx_deriv_real[len(rx_deriv_real)/2::],2)))
            rx_deriv_real = rx_deriv_real.reshape((2, self.mesh.nE)).T
            # Extract the data
            if self.component == "imag":
                rx_deriv_component = 1j * rx_deriv_real
            elif self.component == "real":
                rx_deriv_component = rx_deriv_real.astype(complex)
        else:
            if "xx" in self.orientation:
                ZijN_uV = (
                    self._sDiag(self._hy_py) * self._ex_px_u(v)
                    + self._sDiag(self._ex_px) * self._hy_py_u(v)
                    - self._sDiag(self._ex_py) * self._hy_px_u(v)
                    - self._sDiag(self._hy_px) * self._ex_py_u(v)
                )
            elif "xy" in self.orientation:
                ZijN_uV = (
                    -self._sDiag(self._hx_py) * self._ex_px_u(v)
                    - self._sDiag(self._ex_px) * self._hx_py_u(v)
                    + self._sDiag(self._ex_py) * self._hx_px_u(v)
                    + self._sDiag(self._hx_px) * self._ex_py_u(v)
                )
            elif "yx" in self.orientation:
                ZijN_uV = (
                    self._sDiag(self._hy_py) * self._ey_px_u(v)
                    + self._sDiag(self._ey_px) * self._hy_py_u(v)
                    - self._sDiag(self._ey_py) * self._hy_px_u(v)
                    - self._sDiag(self._hy_px) * self._ey_py_u(v)
                )
            elif "yy" in self.orientation:
                ZijN_uV = (
                    -self._sDiag(self._hx_py) * self._ey_px_u(v)
                    - self._sDiag(self._ey_px) * self._hx_py_u(v)
                    + self._sDiag(self._ey_py) * self._hx_px_u(v)
                    + self._sDiag(self._hx_px) * self._ey_py_u(v)
                )

            Zij = self.eval(src, self.mesh, self.f, True)
            # Calculate the complex derivative
            rx_deriv_real = self._Hd * (ZijN_uV - self._sDiag(Zij) * self._Hd_uV(v))
            rx_deriv_component = np.array(getattr(rx_deriv_real, self.component))

        return rx_deriv_component


class Point_tipper3D(BaseRxNSEM_Point):
    """
    Natural source 3D tipper receiver base class

    :param numpy.ndarray locs: receiver locations (ie. :code:`np.r_[x,y,z]`)
    :param string orientation: receiver orientation 'x', 'y' or 'z'
    :param string component: real or imaginary component 'real' or 'imag'
    """

    def __init__(self, locs, orientation=None, component=None):

        BaseRxNSEM_Point.__init__(
            self, locs, orientation=orientation, component=component
        )

    def eval(self, src, mesh, f, return_complex=False):
        """
        Project the fields to natural source data.

        :param SrcNSEM src: The source of the fields to project
        :param discretize.TensorMesh mesh: Mesh defining the topology of the problem
        :param FieldsNSEM f: Natural source fields object to project
        :rtype: numpy.array
        :return: Evaluated component of the impedance data
        """
        # NOTE: Maybe set this as a property
        self.src = src
        self.mesh = mesh
        self.f = f

        if "zx" in self.orientation:
            Tij = -self._hy_px * self._hz_py + self._hy_py * self._hz_px
        if "zy" in self.orientation:
            Tij = self._hx_px * self._hz_py - self._hx_py * self._hz_px
        rx_eval_complex = self._Hd * Tij

        # Return the full impedance
        if return_complex:
            return rx_eval_complex
        return getattr(rx_eval_complex, self.component)

    def evalDeriv(self, src, mesh, f, v, adjoint=False):
        """
        The derivative of the projection wrt u

        :param SimPEG.EM.NSEM.SrcNSEM src: NSEM source
        :param discretize.TensorMesh mesh: Mesh defining the topology of the problem
        :param SimPEG.EM.NSEM.FieldsNSEM f: NSEM fields object of the source
        :param numpy.ndarray v: Random vector of size
        :rtype: numpy.array
        :return: Calculated derivative (nD,) (adjoint=False) and (nP,2) (adjoint=True)
            for both polarizations
        """
        self.src = src
        self.mesh = mesh
        self.f = f

        if adjoint:
            if "zx" in self.orientation:
                Tij = self._sDiag(
                    self._aHd
                    * (
                        -self._sDiag(self._ahz_py) * self._ahy_px
                        + self._sDiag(self._ahz_px) * self._ahy_py
                    )
                )

                def TijN_uV(x):
                    return (
                        -self._ahz_py_u(self._sDiag(self._ahy_px) * x)
                        - self._ahy_px_u(self._sDiag(self._ahz_py) * x)
                        + self._ahy_py_u(self._sDiag(self._ahz_px) * x)
                        + self._ahz_px_u(self._sDiag(self._ahy_py) * x)
                    )

            elif "zy" in self.orientation:
                Tij = self._sDiag(
                    self._aHd
                    * (
                        self._sDiag(self._ahz_py) * self._ahx_px
                        - self._sDiag(self._ahz_px) * self._ahx_py
                    )
                )

                def TijN_uV(x):
                    return (
                        self._ahx_px_u(self._sDiag(self._ahz_py) * x)
                        + self._ahz_py_u(self._sDiag(self._ahx_px) * x)
                        - self._ahx_py_u(self._sDiag(self._ahz_px) * x)
                        - self._ahz_px_u(self._sDiag(self._ahx_py) * x)
                    )

            # Calculate the complex derivative
            rx_deriv_real = TijN_uV(self._aHd * v) - self._aHd_uV(Tij.T * self._aHd * v)
            # NOTE: Need to reshape the output to go from 2*nU array to a (nU,2) matrix for each polarization
            # rx_deriv_real = np.hstack((mkvc(rx_deriv_real[:len(rx_deriv_real)/2],2),mkvc(rx_deriv_real[len(rx_deriv_real)/2::],2)))
            rx_deriv_real = rx_deriv_real.reshape((2, self.mesh.nE)).T
            # Extract the data
            if self.component == "imag":
                rx_deriv_component = 1j * rx_deriv_real
            elif self.component == "real":
                rx_deriv_component = rx_deriv_real.astype(complex)
        else:
            if "zx" in self.orientation:
                TijN_uV = (
                    -self._sDiag(self._hy_px) * self._hz_py_u(v)
                    - self._sDiag(self._hz_py) * self._hy_px_u(v)
                    + self._sDiag(self._hy_py) * self._hz_px_u(v)
                    + self._sDiag(self._hz_px) * self._hy_py_u(v)
                )
            elif "zy" in self.orientation:
                TijN_uV = (
                    self._sDiag(self._hz_py) * self._hx_px_u(v)
                    + self._sDiag(self._hx_px) * self._hz_py_u(v)
                    - self._sDiag(self._hx_py) * self._hz_px_u(v)
                    - self._sDiag(self._hz_px) * self._hx_py_u(v)
                )
            Tij = self.eval(src, mesh, f, True)
            # Calculate the complex derivative
            rx_deriv_complex = self._Hd * (TijN_uV - self._sDiag(Tij) * self._Hd_uV(v))
            rx_deriv_component = np.array(getattr(rx_deriv_complex, self.component))

        return rx_deriv_component



#  Copyright (c) 2021 Mira Geoscience Ltd.
#
#  This file is part of geoapps.
#
#  geoapps is distributed under the terms and conditions of the MIT License
#  (see LICENSE file at the root of this source code package).

import discretize
import ipywidgets as widgets
import matplotlib.pyplot as plt
import matplotlib.tri as tri
import numpy as np
from geoh5py.groups import ContainerGroup
from geoh5py.objects import Surface
from geoh5py.workspace import Workspace

from geoapps.selection import ObjectDataSelection
from geoapps.simpegPF import PF, Maps, Utils
from geoapps.utils import RectangularBlock, tensor_2_block_model


def block_model_widget(h5file, inducing_field="50000, 90, 0"):
    workspace = Workspace(h5file)
    names = list(workspace.list_objects_name.values())

    def set_axes_equal(ax):
        """
        Source:
        https://stackoverflow.com/questions/13685386/

        Make axes of 3D plot have equal scale so that spheres appear as spheres,
        cubes as cubes, etc..  This is one possible solution to Matplotlib's
        ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

        Input
          ax: a matplotlib axis, e.g., as output from plt.gca().
        """

        x_limits = ax.get_xlim3d()
        y_limits = ax.get_ylim3d()
        z_limits = ax.get_zlim3d()

        x_range = abs(x_limits[1] - x_limits[0])
        x_middle = np.mean(x_limits)
        y_range = abs(y_limits[1] - y_limits[0])
        y_middle = np.mean(y_limits)
        z_range = abs(z_limits[1] - z_limits[0])
        z_middle = np.mean(z_limits)

        # The plot bounding box is a sphere in the sense of the infinity
        # norm, hence I call half the max range the plot radius.
        plot_radius = 0.5 * max([x_range, y_range, z_range])

        ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
        ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
        ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])

    def plot_layout(objects, data, blocks, dip, azimuth, update):

        if workspace.get_entity(objects):
            obj = workspace.get_entity(objects)[0]
        else:
            return

        if obj.get_data(data):
            obs = obj.get_data(data)[0]
        else:
            obs = None

        plt.figure(figsize=(10, 12))
        axs = plt.subplot(projection="3d")
        axs.view_init(dip, ((450 - azimuth) % 360) + 180)
        if getattr(obj, "vertices", None) is not None:

            if isinstance(getattr(obs, "values", None), np.float):
                values = obs.values
            else:
                values = None

            axs.scatter(
                obj.vertices[:, 0],
                obj.vertices[:, 1],
                obj.vertices[:, 2],
                s=4,
                c=values,
                cmap="Spectral_r",
            )
        elif getattr(obj, "centroids", None) is not None:

            if isinstance(obs.values[0], np.float):
                values = obs.values
            else:
                values = None

            axs.scatter(
                obj.centroids[:, 0],
                obj.centroids[:, 1],
                obj.centroids[:, 2],
                s=4,
                c=values,
                cmap="Spectral_r",
            )

        for block in blocks:
            param = blocks_widgets[block].children
            block = RectangularBlock(
                dip=param[0].value,
                azimuth=param[1].value,
                center=[param[2].value, param[3].value, param[4].value],
                length=param[5].value,
                width=param[6].value,
                depth=param[7].value,
            )
            axs.plot_trisurf(
                block.vertices[:, 0],
                block.vertices[:, 1],
                block.vertices[:, 2],
                triangles=block.triangles,
            )

        set_axes_equal(axs)

    object_selection = ObjectDataSelection(h5file=h5file, interactive=True)

    xyz = []
    for obj_name in object_selection.objects.options:
        if workspace.get_entity(obj_name):
            obj = workspace.get_entity(obj_name)[0]
            if getattr(obj, "vertices", None) is not None:
                xyz += [obj.vertices]
            else:
                xyz += [obj.centroids]

    xyz = np.vstack(xyz)

    azimuth = widgets.FloatSlider(
        min=-180,
        max=180,
        value=0,
        step=5,
        description="Camera azimuth",
        continuous_update=False,
    )

    dip = widgets.FloatSlider(
        min=-90,
        max=90,
        value=15,
        step=5,
        description="Camera dip",
        continuous_update=False,
    )

    # Pre-build a list of blocks
    update = widgets.ToggleButton(value=False)

    def update_view(_):
        if update.value:
            update.value = False
        else:
            update.value = True

    blocks_widgets = {}
    for ii in range(10):
        blocks_widgets[ii + 1] = widgets.VBox(
            [
                widgets.FloatSlider(
                    min=-90,
                    max=90,
                    value=0,
                    description="Dip",
                    continuous_update=False,
                ),
                widgets.FloatSlider(
                    min=-180,
                    max=180,
                    value=0,
                    description="Strike",
                    continuous_update=False,
                ),
                widgets.FloatSlider(
                    min=xyz[:, 0].min(),
                    max=xyz[:, 0].max(),
                    value=np.mean(xyz[:, 0]),
                    description="X center:",
                    continuous_update=False,
                ),
                widgets.FloatSlider(
                    min=xyz[:, 1].min(),
                    max=xyz[:, 1].max(),
                    value=np.mean(xyz[:, 1]),
                    description="Y center:",
                    continuous_update=False,
                ),
                widgets.FloatSlider(
                    min=-1000,
                    max=xyz[:, 2].max(),
                    value=np.mean(xyz[:, 2]) - 500,
                    description="Z center:",
                    continuous_update=False,
                ),
                widgets.FloatSlider(
                    min=10,
                    max=10000,
                    value=1000,
                    description="Length:",
                    continuous_update=False,
                ),
                widgets.FloatSlider(
                    min=10,
                    max=10000,
                    value=1000,
                    description="Width:",
                    continuous_update=False,
                ),
                widgets.FloatSlider(
                    min=10,
                    max=10000,
                    value=1000,
                    description="Depth:",
                    continuous_update=False,
                ),
                widgets.FloatSlider(
                    min=0,
                    max=1.0,
                    value=0.01,
                    step=0.001,
                    description="Susceptibility:",
                    continuous_update=False,
                ),
            ]
        )

        for child in blocks_widgets[ii + 1].children:
            child.observe(update_view)

    block_add = widgets.SelectMultiple(
        options=list(blocks_widgets.keys()), description="Select blocks"
    )

    def block_add_update(_):
        block_list.options = block_add.value

    block_add.observe(block_add_update)

    block_list = widgets.Dropdown(
        options=[],
    )

    block_panel = widgets.VBox(
        [
            block_list,
        ]
    )

    def block_param_update(_):
        if block_list.value is not None:
            block_panel.children = [block_list, blocks_widgets[block_list.value]]
        else:
            block_panel.children = [
                block_list,
            ]

    block_list.observe(block_param_update)

    interactive_plot = widgets.interactive_output(
        plot_layout,
        {
            "objects": object_selection.objects,
            "data": object_selection.data,
            "blocks": block_add,
            "azimuth": azimuth,
            "dip": dip,
            "update": update,
        },
    )

    sim_name = widgets.Text(
        value="Simulation_",
        description="Name:",
        disabled=False,
    )

    core_cell_size = widgets.Text(
        value="100, 100, 100",
        description="Model discretization (m)",
        disabled=False,
    )

    topography = widgets.Dropdown(
        options=[None] + names,
        description="Topography",
    )

    # Pre-build a list of blocks
    export_ga = widgets.ToggleButton(
        value=False, description="Export to GA", icon="check"
    )

    def export_trigger(_):
        if getattr(export_ga, "data", None) is not None and export_ga.value:

            # Create a group
            out_group = ContainerGroup.create(workspace, name=sim_name.value)

            obj = workspace.get_entity(object_selection.objects.value)[0]
            obj_out = obj.copy(parent=out_group)

            # Export data on Points
            nC = len(components.value)

            for ind, comp in enumerate(components.value):
                obj_out.add_data(
                    {sim_name.value + "_" + comp: {"values": export_ga.data[ind::nC]}}
                )

            # Export mesh and model
            tensor_2_block_model(
                workspace,
                export_ga.mesh,
                name=sim_name.value + "mesh",
                parent=out_group,
                data={sim_name.value + "_model": export_ga.model},
            )

            # Export blocks as surfaces
            for block in export_ga.blocks:
                Surface.create(
                    workspace,
                    vertices=block.vertices,
                    cells=block.triangles,
                    parent=out_group,
                )
            export_ga.value = False

    export_ga.observe(export_trigger)

    plot_now = widgets.ToggleButton(
        value=True, description="Plot simulation", icon="check"
    )

    def run_simulation(_):

        if forward.value:
            obj = workspace.get_entity(object_selection.objects.value)[0]

            if getattr(obj, "vertices", None) is not None:
                xyz = obj.vertices
            else:
                xyz = obj.centroids

            nodes = []
            out_blocks = []
            for block in block_add.value:
                param = blocks_widgets[block].children
                block = RectangularBlock(
                    dip=param[0].value,
                    azimuth=param[1].value,
                    center=[param[2].value, param[3].value, param[4].value],
                    length=param[5].value,
                    width=param[6].value,
                    depth=param[7].value,
                    susc=param[8].value,
                )
                out_blocks.append(block)
                nodes.append(block.vertices)

            h = np.asarray(core_cell_size.value.split(",")).astype(float).tolist()
            paddings = [[h[0], h[0]], [h[1], h[1]], [h[2], h[2]]]

            # Create a mesh
            mesh = discretize.utils.meshutils.mesh_builder_xyz(
                np.vstack(nodes), h, padding_distance=paddings
            )

            # Create a model
            model = np.zeros(mesh.nC)
            for block in out_blocks:
                ind = Utils.ModelBuilder.PolygonInd(mesh, block.vertices)
                model[ind] = block.susc

            # Cut air cells
            if topography.value is not None:
                topo = workspace.get_entity(topography.value)[0]
                ind = Utils.modelutils.surface2ind_topo(mesh, topo.vertices)
                model[ind == False] = -1

            # Create a problem
            active = model > 0
            nC = int(active.sum())

            if survey_type.value == "Magnetics":
                rxLoc = PF.BaseMag.RxObs(xyz)
                srcField = PF.BaseMag.SrcField(
                    [rxLoc],
                    param=np.asarray(inducing_field.value.split(","))
                    .astype(float)
                    .tolist(),
                )
                survey = PF.BaseMag.LinearSurvey(srcField, components=components.value)
                prob = PF.Magnetics.MagneticIntegral(
                    mesh,
                    chiMap=Maps.IdentityMap(nP=nC),
                    actInd=active,
                    forwardOnly=True,
                    verbose=False,
                )
                prob.pair(survey)
                d = prob.fields(model[active])

            else:
                rxLoc = PF.BaseGrav.RxObs(xyz)
                srcField = PF.BaseGrav.SrcField([rxLoc])
                survey = PF.BaseGrav.LinearSurvey(srcField, components=components.value)
                prob = PF.Gravity.GravityIntegral(
                    mesh,
                    rhoMap=Maps.IdentityMap(nP=nC),
                    actInd=active,
                    forwardOnly=True,
                    verbose=False,
                )
                prob.pair(survey)
                d = prob.fields(model[active])

            # Save latest simulation to widget for export
            export_ga.data = d
            export_ga.mesh = mesh
            export_ga.model = model
            export_ga.blocks = out_blocks
        forward.value = False

    # Pre-build a list of blocks
    forward = widgets.ToggleButton(value=False, description="Run forward", icon="check")
    forward.observe(run_simulation)

    forward_panel = widgets.VBox(
        [
            widgets.Label("FORWARD SIMULATION"),
            topography,
            core_cell_size,
            sim_name,
            forward,
            plot_now,
            export_ga,
        ]
    )

    def update_data_options(_):
        if survey_type.value == "Magnetics":
            components.options = ["tmi", "bxx", "bxy", "bxz", "byy", "byz", "bzz"]
            components.value = ["tmi"]
            phys_prop = "Susceptibility SI"
            ranges = [0.0, 1.0]
        else:
            components.options = ["gz", "gxx", "gxy", "gxz", "gyy", "gyz", "gzz"]
            components.value = ["gz"]
            phys_prop = "Residual Density (g/cc)"
            ranges = [-1.0, 1.0]

        for widget in blocks_widgets.values():
            widget.children[-1].description = phys_prop
            widget.children[-1].min = ranges[0]
            widget.children[-1].max = ranges[1]

    components = widgets.SelectMultiple(
        description="Components",
        options=["tmi", "bxx", "bxy", "bxz", "byy", "byz", "bzz"],
        value=["tmi"],
    )

    components_panel = widgets.VBox([components])

    def update_survey_type(_):
        if survey_type.value == "Magnetics":
            survey_type_panel.children = [
                survey_type,
                object_selection.main,
                components_panel,
                inducing_field,
            ]
        else:
            survey_type_panel.children = [
                survey_type,
                object_selection.main,
                components_panel,
            ]

        update_data_options("")

    survey_type = widgets.Dropdown(
        options=["Magnetics", "Gravity"],
        value="Magnetics",
        description="Survey Type:",
    )
    inducing_field = widgets.Text(
        value=inducing_field,
        description="Inducing Field [Amp, Inc, Dec]",
    )
    survey_type_panel = widgets.VBox(
        [survey_type, object_selection.main, components_panel, inducing_field]
    )
    survey_type.observe(update_survey_type)

    def plot_simulation(plot, run):

        objects = object_selection.objects.value
        data = object_selection.data.value

        if plot:
            if workspace.get_entity(objects):
                obj = workspace.get_entity(objects)[0]
                if getattr(obj, "vertices", None) is not None:
                    xyz = obj.vertices
                else:
                    xyz = obj.centroids

                if obj.get_data(data):
                    obs = obj.get_data(data)[0]
                else:
                    return
            else:
                return

            nC = len(components.value)

            if (
                getattr(export_ga, "data", None) is not None
                and export_ga.data.shape[0] == nC * xyz.shape[0]
            ):
                trian = tri.Triangulation(xyz[:, 0], xyz[:, 1])

                plt.figure(figsize=(8, 4 * int(np.ceil(nC / 2) + 1)))
                axs = plt.subplot(int(np.ceil(nC / 2) + 1), 2, 1)
                if isinstance(obs.values[0], np.float):
                    im = axs.tricontourf(
                        trian,
                        obs.values,
                        cmap="Spectral_r",
                        levels=100,
                        vmin=obs.values.min(),
                        vmax=obs.values.max(),
                    )
                    plt.colorbar(im)
                    axs.set_title(data)
                    axs.set_yticklabels([])
                    axs.set_xticklabels([])
                    axs.set_aspect("equal")

                for ind, comp in enumerate(components.value):
                    axs = plt.subplot(int(np.ceil(nC / 2) + 1), 2, ind + 2)
                    im = axs.tricontourf(
                        trian,
                        export_ga.data[ind::nC],
                        cmap="Spectral_r",
                        levels=100,
                        vmin=obs.values.min(),
                        vmax=obs.values.max(),
                    )
                    plt.colorbar(im)
                    axs.set_yticklabels([])
                    axs.set_xticklabels([])
                    axs.set_title(comp)
                    axs.set_aspect("equal")

    return widgets.VBox(
        [
            widgets.HBox(
                [
                    widgets.VBox([survey_type_panel, forward_panel]),
                    widgets.VBox([block_add, block_panel]),
                ]
            ),
            widgets.HBox(
                [
                    widgets.VBox(
                        [
                            azimuth,
                            dip,
                            interactive_plot,
                        ]
                    ),
                    widgets.interactive_output(
                        plot_simulation, {"plot": plot_now, "run": forward}
                    ),
                ]
            ),
        ]
    )



#  Copyright (c) 2021 Mira Geoscience Ltd.
#
#  This file is part of geoapps.
#
#  geoapps is distributed under the terms and conditions of the MIT License
#  (see LICENSE file at the root of this source code package).



#  Copyright (c) 2021 Mira Geoscience Ltd.
#
#  This file is part of geoapps.
#
#  geoapps is distributed under the terms and conditions of the MIT License
#  (see LICENSE file at the root of this source code package).

import ipywidgets as widgets
import numpy as np
from geoh5py.data import FloatData, IntegerData, ReferencedData
from geoh5py.objects.object_base import ObjectBase
from geoh5py.workspace import Workspace
from ipywidgets import Dropdown, FloatText, SelectMultiple, VBox

from geoapps import utils
from geoapps.base import BaseApplication


class ObjectDataSelection(BaseApplication):
    """
    Application to select an object and corresponding data
    """

    defaults = {}

    def __init__(self, **kwargs):
        kwargs = self.apply_defaults(**kwargs)

        self._add_groups = False
        self._find_label = []
        self._object_types = []
        self._select_multiple = False

        super().__init__(**kwargs)
        self.data_panel = VBox([self.objects, self.data])
        self.update_data_list(None)
        self._main = self.data_panel

    @property
    def add_groups(self):
        """
        bool: Add data groups to the list of data choices
        """
        return self._add_groups

    @add_groups.setter
    def add_groups(self, value):
        assert isinstance(value, (bool, str)), "add_groups must be of type bool"
        self._add_groups = value

    @property
    def data(self):
        """
        Data selector
        """
        if getattr(self, "_data", None) is None:
            if self.select_multiple:
                self._data = SelectMultiple(
                    description="Data: ",
                )
            else:
                self._data = Dropdown(
                    description="Data: ",
                )
        return self._data

    @data.setter
    def data(self, value):
        assert isinstance(
            value, (Dropdown, SelectMultiple)
        ), f"'Objects' must be of type {Dropdown} or {SelectMultiple}"
        self._data = value

    @property
    def objects(self):
        """
        Object selector
        """
        if getattr(self, "_objects", None) is None:
            self.objects = Dropdown(description="Object:", options=[""])

        return self._objects

    @objects.setter
    def objects(self, value):
        assert isinstance(value, Dropdown), f"'Objects' must be of type {Dropdown}"
        self._objects = value
        self._objects.observe(self.update_data_list, names="value")

    @property
    def object_types(self):
        """
        Entity type
        """
        if getattr(self, "_object_types", None) is None:
            self._object_types = []

        return self._object_types

    @object_types.setter
    def object_types(self, entity_types):
        if not isinstance(entity_types, list):
            entity_types = [entity_types]

        for entity_type in entity_types:
            assert issubclass(
                entity_type, ObjectBase
            ), f"Provided object_types must be instances of {ObjectBase}"

        self._object_types = tuple(entity_types)

    @property
    def find_label(self):
        """
        Object selector
        """
        if getattr(self, "_find_label", None) is None:
            return []

        return self._find_label

    @find_label.setter
    def find_label(self, values):
        """
        Object selector
        """
        if not isinstance(values, list):
            values = [values]

        for value in values:
            assert isinstance(
                value, str
            ), f"Labels to find must be strings. Value {value} of type {type(value)} provided"
        self._find_label = values

    @property
    def select_multiple(self):
        """
        bool: ALlow to select multiple data
        """
        if getattr(self, "_select_multiple", None) is None:
            self._select_multiple = False

        return self._select_multiple

    @select_multiple.setter
    def select_multiple(self, value):
        if getattr(self, "_data", None) is not None:
            options = self._data.options
        else:
            options = []

        self._select_multiple = value

        if value:
            self._data = SelectMultiple(description="Data: ", options=options)
        else:
            self._data = Dropdown(description="Data: ", options=options)

    @property
    def workspace(self):
        """
        Target geoh5py workspace
        """
        if (
            getattr(self, "_workspace", None) is None
            and getattr(self, "_h5file", None) is not None
        ):
            self.workspace = Workspace(self.h5file)
        return self._workspace

    @workspace.setter
    def workspace(self, workspace):
        assert isinstance(workspace, Workspace), f"Workspace must of class {Workspace}"
        self._workspace = workspace
        self._h5file = workspace.h5file

        # Refresh the list of objects
        self.update_objects_list()

    def get_selected_entities(self):
        """
        Get entities from an active geoh5py Workspace
        """
        if getattr(self, "_workspace", None) is not None and self._workspace.get_entity(
            self.objects.value
        ):
            for entity in self._workspace.get_entity(self.objects.value):
                if isinstance(entity, ObjectBase):
                    obj = entity

            if isinstance(self.data, Dropdown):
                values = [self.data.value]
            else:
                values = self.data.value

            data = []
            for value in values:
                if obj.get_data(value):
                    data += obj.get_data(value)

                elif any([pg.name == value for pg in obj.property_groups]):
                    data += [
                        self.workspace.get_entity(prop)[0]
                        for prop in obj.get_property_group(value).properties
                    ]

            return obj, data
        else:
            return None, None

    def update_data_list(self, _):
        self.refresh.value = False
        if getattr(self, "_workspace", None) is not None and self._workspace.get_entity(
            self.objects.value
        ):
            for entity in self._workspace.get_entity(self.objects.value):
                if isinstance(entity, ObjectBase):
                    obj = entity

            if getattr(obj, "get_data_list", None) is None:
                return

            options = [""]

            if (self.add_groups or self.add_groups == "only") and obj.property_groups:
                options = (
                    options
                    + ["-- Groups --"]
                    + [p_g.name for p_g in obj.property_groups]
                )

            if self.add_groups != "only":
                data_list = obj.get_data_list()
                options = (
                    options
                    + ["--- Channels ---"]
                    + [
                        obj.get_data(uid)[0].name
                        for uid in data_list
                        if isinstance(obj.get_data(uid)[0], (IntegerData, FloatData))
                    ]
                    + ["Z"]
                )

            value = self.data.value
            self.data.options = options

            if self.select_multiple and any([val in options for val in value]):
                self.data.value = [val for val in value if val in options]
            elif value in options:
                self.data.value = value
            elif self.find_label:
                self.data.value = utils.find_value(self.data.options, self.find_label)
        else:
            self.data.options = []

        self.refresh.value = True

    def update_objects_list(self):
        if getattr(self, "_workspace", None) is not None:
            value = self.objects.value

            if len(self.object_types) > 0:
                options = [""] + [
                    obj.name
                    for obj in self._workspace.all_objects()
                    if isinstance(obj, self.object_types)
                ]
            else:
                options = [""] + list(self._workspace.list_objects_name.values())

            if value in options:  # Silent update
                self.objects.unobserve(self.update_data_list, names="value")
                self.objects.options = options
                self.objects.value = value
                self._objects.observe(self.update_data_list, names="value")
            else:
                self.objects.options = options


class LineOptions(ObjectDataSelection):
    """
    Unique lines selection from selected data channel
    """

    defaults = {"find_label": "line"}

    def __init__(self, **kwargs):
        kwargs = self.apply_defaults(**kwargs)

        self._multiple_lines = None

        super().__init__(**kwargs)

        if "objects" in kwargs.keys() and isinstance(kwargs["objects"], Dropdown):
            self._objects.observe(self.update_data_list, names="value")

        self._data.observe(self.update_line_list, names="value")
        self.update_data_list(None)
        self.update_line_list(None)

        self._main = VBox([self._data, self.lines])
        self._data.description = "Lines field"

    @property
    def lines(self):
        """
        Widget.SelectMultiple or Widget.Dropdown
        """
        if getattr(self, "_lines", None) is None:
            if self.multiple_lines:
                self._lines = widgets.SelectMultiple(
                    description="Select lines:",
                )
            else:
                self._lines = widgets.Dropdown(
                    description="Select line:",
                )

        return self._lines

    @property
    def multiple_lines(self):
        if getattr(self, "_multiple_lines", None) is None:
            self._multiple_lines = True

        return self._multiple_lines

    @multiple_lines.setter
    def multiple_lines(self, value):
        assert isinstance(
            value, bool
        ), f"'multiple_lines' property must be of type {bool}"
        self._multiple_lines = value

    def update_line_list(self, _):
        _, data = self.get_selected_entities()
        if data and getattr(data[0], "values", None) is not None:
            if isinstance(data[0], ReferencedData):
                self.lines.options = [""] + list(data[0].value_map.map.values())
            else:
                self.lines.options = [""] + np.unique(data[0].values).tolist()


class TopographyOptions(ObjectDataSelection):
    """
    Define the topography used by the inversion
    """

    def __init__(self, **kwargs):
        self.find_label = ["topo", "dem", "dtm", "elevation", "Z"]
        self._offset = FloatText(description="Vertical offset (+ve up)")
        self._constant = FloatText(
            description="Elevation (m)",
        )

        super().__init__(**kwargs)

        self.objects.value = utils.find_value(self.objects.options, self.find_label)
        self.option_list = {
            "Object": self.main,
            "Relative to Sensor": self.offset,
            "Constant": self.constant,
            "None": widgets.Label("No topography"),
        }
        self._options = widgets.RadioButtons(
            options=["Object", "Relative to Sensor", "Constant"],
            description="Define by:",
        )

        def update_options(_):
            self.update_options()

        self.options.observe(update_options)
        self._main = VBox([self.options, self.option_list[self.options.value]])

    @property
    def panel(self):
        return self._panel

    @property
    def constant(self):
        return self._constant

    @property
    def offset(self):
        return self._offset

    @property
    def options(self):
        return self._options

    def update_options(self):
        self._main.children = [
            self.options,
            self.option_list[self.options.value],
        ]



# import numpy as np
# from scipy.constants import mu_0
# from .DigFilter import EvalDigitalFilt
# from RTEfun import rTEfun

# def HzKernel_layer(lamda, f, nlay, sig, chi, depth, h, z, flag):

#     """

#         Kernel for vertical magnetic component (Hz) due to vertical magnetic
#         diopole (VMD) source in (kx,ky) domain

#     """
#     u0 = lamda
#     rTE, M00, M01, M10, M11 = rTEfun(nlay, f, lamda, sig, chi, depth)

#     if flag=='secondary':
#         # Note
#         # Here only computes secondary field.
#         # I am not sure why it does not work if we add primary term.
#         # This term can be analytically evaluated, where h = 0.

#         kernel = 1/(4*np.pi)*(rTE*np.exp(-u0*(z+h)))*lamda**3/u0

#     else:
#         kernel = 1/(4*np.pi)*(np.exp(u0*(z-h))+ rTE*np.exp(-u0*(z+h)))*lamda**3/u0

#     return  kernel

# def HzkernelCirc_layer(lamda, f, nlay, sig, chi, depth, h, z, I, a, flag):

#     """

#         Kernel for vertical magnetic component (Hz) at the center
#         due to circular loop source in (kx,ky) domain

#         .. math::

#             H_z = \\frac{Ia}{2} \int_0^{\infty} [e^{-u_0|z+h|} + r_{TE}e^{u_0|z-h|}] \\frac{\lambda^2}{u_0} J_1(\lambda a)] d \lambda

#     """

#     w = 2*np.pi*f
#     rTE = np.zeros(lamda.size, dtype=complex)
#     u0 = lamda
#     rTE, M00, M01, M10, M11 = rTEfun(nlay, f, lamda, sig, chi, depth)

#     if flag == 'secondary':
#         kernel = I*a*0.5*(rTE*np.exp(-u0*(z+h)))*lamda**2/u0
#     else:
#         kernel = I*a*0.5*(np.exp(u0*(z-h))+rTE*np.exp(-u0*(z+h)))*lamda**2/u0

#     return  kernel

# TODO: Get rid of below two functions and put in in main class
# def HzFreq_layer(nlay, sig, chi, depth, f, z, h, r, flag, YBASE, WT0):
#     """

#     """
#     nfreq = np.size(f)
#     HzFHT = np.zeros(nfreq, dtype = complex)
#     for ifreq in range(nfreq):

#         kernel = lambda x: HzKernel_layer(x, f[ifreq], nlay, sig, chi, depth, h, z, flag)
#         HzFHT[ifreq] = EvalDigitalFilt(YBASE, WT0, kernel, r)

#     return HzFHT

# def HzCircFreq_layer(nlay, sig, chi, depth, f, z, h, I, a, flag, YBASE, WT1):

#     """

#     """
#     nfreq = np.size(f)
#     HzFHT = np.zeros(nfreq, dtype = complex)
#     for ifreq in range(nfreq):

#         kernel = lambda x: HzkernelCirc_layer(x, f[ifreq], nlay, sig, chi, depth, h, z, I, a, flag)
#         HzFHT[ifreq] = EvalDigitalFilt(YBASE, WT1, kernel, a)

#     return HzFHT



#  Copyright (c) 2021 Mira Geoscience Ltd.
#
#  This file is part of geoapps.
#
#  geoapps is distributed under the terms and conditions of the MIT License
#  (see LICENSE file at the root of this source code package).


import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
from geoh5py.objects import Curve, Grid2D, Points, Surface
from ipywidgets import FloatSlider, FloatText, HBox, Label, Layout, ToggleButton, VBox

from geoapps.plotting import plot_plan_data_selection
from geoapps.selection import ObjectDataSelection
from geoapps.utils import input_string_2_float, rotate_xy


class PlotSelection2D(ObjectDataSelection):
    """
    Application for selecting data in 2D plan map view
    """

    defaults = {
        "h5file": "../../assets/FlinFlon.geoh5",
        "objects": "Gravity_Magnetics_drape60m",
        "data": "Airborne_TMI",
    }

    def __init__(self, **kwargs):
        self.figure = None
        self.axis = None
        self.indices = None
        self.collections = []
        self._azimuth = FloatSlider(
            min=-90,
            max=90,
            value=0,
            step=5,
            description="Azimuth",
            continuous_update=False,
        )
        self._center_x = FloatSlider(
            min=-100,
            max=100,
            step=10,
            description="Easting",
            continuous_update=False,
        )
        self._center_y = FloatSlider(
            min=-100,
            max=100,
            step=10,
            description="Northing",
            continuous_update=False,
            orientation="vertical",
        )
        self._contours = widgets.Text(
            value="",
            description="Contours",
            disabled=False,
            continuous_update=False,
        )
        self._data_count = Label("Data Count: 0")
        self._resolution = FloatText(
            description="Grid Resolution (m)", style={"description_width": "initial"}
        )
        self._width = FloatSlider(
            min=0,
            max=100,
            step=10,
            value=1000,
            description="Width",
            continuous_update=False,
        )
        self._height = FloatSlider(
            min=0,
            max=100,
            step=10,
            value=1000,
            description="Height",
            continuous_update=False,
            orientation="vertical",
        )
        self._zoom_extent = ToggleButton(
            value=True,
            description="Zoom on selection",
            icon="check",
        )

        def set_bounding_box(_):
            self.set_bounding_box()

        self.highlight_selection = None

        super().__init__(**self.apply_defaults(**kwargs))

        self.objects.observe(set_bounding_box, names="value")
        self.set_bounding_box()

        self.window_plot = widgets.interactive_output(
            self.plot_selection,
            {
                "data_name": self.data,
                "resolution": self.resolution,
                "center_x": self.center_x,
                "center_y": self.center_y,
                "width": self.width,
                "height": self.height,
                "azimuth": self.azimuth,
                "zoom_extent": self.zoom_extent,
                "contours": self.contours,
                "refresh": self.refresh,
            },
        )

        self.window_selection = VBox(
            [
                VBox([self.resolution, self.data_count]),
                HBox(
                    [
                        self.center_y,
                        self.height,
                        VBox(
                            [
                                self.width,
                                self.center_x,
                                self.window_plot,
                                self.azimuth,
                                self.zoom_extent,
                            ]
                        ),
                    ],
                    layout=Layout(align_items="center"),
                ),
            ]
        )
        self._main = VBox([self.main, self.window_selection])

        if "window" in kwargs.keys():
            self.refresh.value = False
            self.__populate__(**kwargs["window"])
            self.refresh.value = True

    @property
    def azimuth(self):
        """
        :obj:`ipywidgets.FloatSlider`: Rotation angle of the selection box.
        """
        return self._azimuth

    @property
    def center_x(self):
        """
        :obj:`ipywidgets.FloatSlider`: Easting position of the selection box.
        """
        return self._center_x

    @property
    def center_y(self):
        """
        :obj:`ipywidgets.FloatSlider`: Northing position of the selection box.
        """
        return self._center_y

    @property
    def contours(self):
        """
        :obj:`ipywidgets.widgets.Text` String defining sets of contours.
        Contours can be defined over an interval `50:200:10` and/or at a fix value `215`.
        Any combination of the above can be used:
        50:200:10, 215 => Contours between values 50 and 200 every 10, with a contour at 215.
        """
        return self._contours

    @property
    def data_count(self):
        """
        :obj:`ipywidgets.Label`: Data counter included in the selection box.
        """
        return self._data_count

    @property
    def height(self):
        """
        :obj:`ipywidgets.FloatSlider`: Height (m) of the selection box
        """
        return self._height

    @property
    def resolution(self):
        """
        :obj:`ipywidgets.FloatText`: Minimum data separation (m)
        """
        return self._resolution

    @property
    def width(self):
        """
        :obj:`ipywidgets.FloatSlider`: Width (m) of the selection box
        """
        return self._width

    @property
    def zoom_extent(self):
        """
        :obj:`ipywidgets.ToggleButton`: Set plotting limits to the selection box
        """
        return self._zoom_extent

    def plot_selection(
        self,
        data_name,
        resolution,
        center_x,
        center_y,
        width,
        height,
        azimuth,
        zoom_extent,
        contours,
        refresh,
    ):
        if not refresh:
            return

        # Parse the contours string
        contours = input_string_2_float(contours)

        entity, _ = self.get_selected_entities()
        if entity is None:
            return
        data_obj = None

        if hasattr(self, "plotting_data"):
            data_channel = self.plotting_data
        else:
            if self.select_multiple and data_name:
                data_channel = data_name[0]
            else:
                data_channel = data_name

        if entity.get_data(data_channel):
            data_obj = entity.get_data(data_channel)[0]
        elif data_channel == "Z":
            data_obj = "Z"

        if isinstance(entity, (Grid2D, Surface, Points, Curve)):

            self.figure = plt.figure(figsize=(10, 10))
            self.axis = plt.subplot()
            corners = np.r_[
                np.c_[-1.0, -1.0],
                np.c_[-1.0, 1.0],
                np.c_[1.0, 1.0],
                np.c_[1.0, -1.0],
                np.c_[-1.0, -1.0],
            ]
            corners[:, 0] *= width / 2
            corners[:, 1] *= height / 2
            corners = rotate_xy(corners, [0, 0], -azimuth)
            self.axis.plot(corners[:, 0] + center_x, corners[:, 1] + center_y, "k")
            self.axis, _, ind_filter, _, contour_set = plot_plan_data_selection(
                entity,
                data_obj,
                **{
                    "axis": self.axis,
                    "resolution": resolution,
                    "window": {
                        "center": [center_x, center_y],
                        "size": [width, height],
                        "azimuth": azimuth,
                    },
                    "zoom_extent": zoom_extent,
                    "resize": True,
                    "contours": contours,
                    "highlight_selection": self.highlight_selection,
                    "collections": self.collections,
                },
            )

            self.indices = ind_filter
            self.contours.contour_set = contour_set
            self.data_count.value = f"Data Count: {ind_filter.sum()}"

    def set_bounding_box(self):
        # Fetch vertices in the project
        lim_x = [1e8, -1e8]
        lim_y = [1e8, -1e8]

        obj, _ = self.get_selected_entities()
        if isinstance(obj, Grid2D):
            lim_x[0], lim_x[1] = obj.centroids[:, 0].min(), obj.centroids[:, 0].max()
            lim_y[0], lim_y[1] = obj.centroids[:, 1].min(), obj.centroids[:, 1].max()
        elif isinstance(obj, (Points, Curve, Surface)):
            lim_x[0], lim_x[1] = obj.vertices[:, 0].min(), obj.vertices[:, 0].max()
            lim_y[0], lim_y[1] = obj.vertices[:, 1].min(), obj.vertices[:, 1].max()
        else:
            return

        width = lim_x[1] - lim_x[0]
        height = lim_y[1] - lim_y[0]

        self.refresh.value = False
        self.center_x.min = -np.inf
        self.center_x.max = lim_x[1] + width * 0.1
        self.center_x.value = np.mean(lim_x)
        self.center_x.min = lim_x[0] - width * 0.1

        self.center_y.min = -np.inf
        self.center_y.max = lim_y[1] + height * 0.1
        self.center_y.value = np.mean(lim_y)
        self.center_y.min = lim_y[0] - height * 0.1

        self.width.max = width * 1.2
        self.width.value = self.width.max / 2.0
        self.width.min = 0

        self.height.max = height * 1.2
        self.height.min = 0
        self.height.value = self.height.max / 2.0
        self.refresh.value = True
